#!/usr/bin/env python3
"""
Utility methods for tags
"""
import re

import logging
from utils import match

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def get_quad_tag_parts(reftag:str):
    """
    Return a 4-tuple given a tag
    """

    # Split tag into parts
    tag_parts = reftag.split("/")

    if len(tag_parts) < 3 or len(tag_parts) > 4:
        return None, None, None, None

    tag_prefix = tag_parts[0]
    tag_context_name = tag_parts[1]
    tag_value_name = tag_parts[2]
    tag_comparison = "equals"
    if len(tag_parts) == 4:
        tag_comparison = tag_parts[3]

    return tag_prefix, tag_context_name, tag_value_name, tag_comparison

def check_tag_comparison(tag_tuple, compare_value:str, compare_to_key, compare_to_value, callback, callback_config, only_callback = False):
    """
    Using the tag comparison method, check if two key values compare
    """

    tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tag_tuple

    if not only_callback and tag_comparison == "" or tag_comparison == "equals":
        if match.equals(compare_value, compare_to_value):
            return True
    if not only_callback and tag_comparison == "endswith":
        if match.endswith(compare_value, compare_to_value):
            return True
    else:
        if callback is not None:
            return callback(callback_config, tag_tuple, compare_value, compare_to_key, compare_to_value)    

    return False

def get_prefixed_tag(prefix:str, target_key) -> list:
    """
    Return all the tags on a key with the given prefix
    """

    output = []

    for tag in target_key.getTags():
        if tag.startswith(prefix):
            output.append(tag)

    return output


# Return a list tags of all tags matching the regex
def tags_matching_regex(key, regex_str:str, compiled_regex = None):
    """ 
    Returns a list of tags with all tags that match the regular expression
    """

    if compiled_regex is None:
         compiled_regex = re.compile(regex_str)

    tags_list = []

    for tag in key.getTags():
        if compiled_regex.search(tag) is not None:
            tags_list.append(tag)
    
    return tags_list


# def get_values_for_tags(model, ref_tag_list:list, output:dict):
#     """

#     """

#     for tag in ref_tag_list:

#         tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = get_quad_tag_parts(tag)

#         for key_entry, value_entry in find.keys_with_tag(model, tag):
#             if isinstance(value_entry, str):
#                 if match.is_empty(value_entry):
#                     continue
#                 tag_dict_key = tag_data_tag_name + "/" + tag_field_tag_name
#                 if output.get(tag_dict_key) is None:
#                     output[tag_dict_key] = []
#                 if value_entry not in output[tag_dict_key]:
#                     output[tag_dict_key].append(value_entry)
#             else:
#                 logger.warning(f"Reference value '{tag}' did not point to a string")

#     return                

def get_matching_tags(key, prefix, context_name, value_name, comparison):
    """
    Return all tags that match the passed in tag parts
    """

    output = []

    for tag in key.getTags():
        
        tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = get_quad_tag_parts(tag)

        if tag_prefix == prefix or prefix == "":
            if tag_data_tag_name == context_name or context_name == "":
                if tag_field_tag_name == value_name or value_name == "":
                    if tag_comparison == comparison or comparison == "":
                        output.append(tag)

    return output


#!/usr/bin/env python3
"""
Reports on verifier output
"""

import logging
import jsonpickle
from utils.load_yaml import yaml_register_class
from verifiers.verifiers_config import VerifiersConfig

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class VerifiersReport:
    """ Produces a report on the output of the verification action

    """

    def __init__(self, config:VerifiersConfig, display_asset_coverage:bool = True, display_control_coverage:bool = True):
        yaml_register_class(VerifiersReport)

        self.config = config
        self.display_asset_coverage = display_asset_coverage
        self.display_control_coverage = display_control_coverage

    def report(self, verifier_issues_list:list, covered_assets_list:list) -> dict:
        self.verifier_issues_list = verifier_issues_list
        self.covered_assets_list = covered_assets_list

        self.control_coverage = {}
        # Generate the control coverage report consistaing of control, asset storage
        for covered_asset in self.covered_assets_list:
            # Loop through the controls
            for covered_threat in covered_asset.covering_threats:
                for control_key, control_value in covered_threat.threat_controls:
                    if control_value not in self.control_coverage:
                        self.control_coverage[control_value] = {}
                    if covered_asset.storage_location_value not in self.control_coverage[control_value]:
                        self.control_coverage[control_value][covered_asset.storage_location_value] = []
                    if covered_asset.asset_value not in self.control_coverage[control_value][covered_asset.storage_location_value]:
                        self.control_coverage[control_value][covered_asset.storage_location_value].append(covered_asset.asset_value)

    def getIssues(self):
        return self.verifier_issues_list

    def _get_state(self):

        def _covered_assets_sort(value):
            return len(value.covering_threats)

        output = {}
        output['issues'] = self.verifier_issues_list
        if self.display_asset_coverage:
            output['asset-coverage-report'] = sorted(self.covered_assets_list, key=_covered_assets_sort)
        if self.display_control_coverage:
            control_report = []
            for control, asset_list in self.control_coverage.items():
                control_report_entry = {}
                control_report_entry["control"] = control
                control_report_entry["locations"] = []
                for location, asset_list in asset_list.items():
                    location_report_entry = {}
                    location_report_entry["location"] = location
                    location_report_entry["assets"] = asset_list
                    control_report_entry["locations"].append(location_report_entry)
                control_report.append(control_report_entry)    
            output['control-coverage-report'] = control_report
            #output['control-coverage-report'] = self.control_coverage
        return output

    def __getstate__(self):
        """ Used by jsonpickle to state of class to output """
        return self._get_state()

    def tojson(self):
        return jsonpickle.encode(self, unpicklable=False)

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node._get_state())


#!/usr/bin/env python3
"""
Sets the base directory from which all configuration files are loaded
"""

import logging
import os
from pathlib import Path
from storage.git_permanent_repo import GitPermanentStorage
from utils.error import StorageError
from utils import shell
from sh import rm

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

DEFAULT_CONFIG_BASE_DIR = "~/.threatware"
DEFAULT_CONFIG_REPO = "https://github.com/samadhicsec/threatware-config.git"
DEFAULT_CONFIG_REPO_BRANCH = "main"

class ConfigBase:

    install_dir:str
    base_dir:str
    ephemeral_env:bool

    @classmethod
    def init(cls, execution_env):

        cls._set_install_directory()

        dynamic_config = os.environ.get("THREATWARE_CONFIG_DYNAMIC")
        if dynamic_config == "False":
            # Only if this is explicitly set to "False" do we not try to use dynamic configuration
            cls.base_dir = cls.install_dir
            logger.info("Dynamnic configuration disabled.  Using configuration depoyed with code.")
            return

        suggested_base_dir = os.getenv("THREATWARE_CONFIG_DIR")
        if suggested_base_dir is not None:
            suggested_base_dir = os.path.expandvars(os.path.expanduser(suggested_base_dir))
        else:
            suggested_base_dir = os.path.expanduser(DEFAULT_CONFIG_BASE_DIR)

        # Ask the execution environment where the config should be
        exe_env_root_dir, cls.ephemeral_env = execution_env.get_config_base_dir(suggested_base_dir)

        # Check if the directory exists
        if not cls.ephemeral_env and os.path.isdir(exe_env_root_dir):

            # It exists, so we expect all the config to be present
            cls.base_dir = exe_env_root_dir
            logger.info(f"Using configuration in '{exe_env_root_dir}'")
            return

        else:

            # There is no available configuration.  There are 2 options:
            # - download it, but if that fails
            # - use built-in config, shipped with code
            try: 
                cls._download_config(exe_env_root_dir, execution_env, cls.ephemeral_env)
            finally:
                # Check if the directory exists
                if os.path.isdir(exe_env_root_dir):

                    # It exists, so we expect all the config to be present
                    cls.base_dir = exe_env_root_dir
                    logger.info(f"Using configuration in '{exe_env_root_dir}'")

                else:
                    # Still no config directory, use built-in config
                    cls.base_dir = cls.install_dir
                    logger.info("Failed to acquire dynamic configuration.  Using configuration depoyed with code.")

    @classmethod
    def _set_install_directory(cls):
        """
        Sets the install directory.  The install directory is the root directory of the threatware modules.
        """

        # We are going to cheat and use the fact that we know this file has path utils/config.py
        cls.install_dir = str(Path(__file__).absolute().parent.parent)

    @classmethod
    def _download_config(cls, exe_env_root_dir, execution_env, ephemeral_env):
        """ Try to download configuration from a configured location """

        delete_git_dir = False

        config_git_branch = os.getenv("THREATWARE_CONFIG_REPO_BRANCH")

        config_git_repo = os.getenv("THREATWARE_CONFIG_REPO")
        if config_git_repo is None:
            if not ephemeral_env:
                # Use the threatware default
                config_git_repo = DEFAULT_CONFIG_REPO
                config_git_branch = DEFAULT_CONFIG_REPO_BRANCH
                # No one ever needs to push to this repo, so remove the .git directory
                delete_git_dir = True
            else:
                # No point in downloading the default config as no creds will be set, so no actions will work
                # Just return, so the config shipped with code is used
                return

        # Get the parent and name of the directory to store the config in    
        exe_env_root_path = Path(exe_env_root_dir)
        exe_env_root_path_parent = str(exe_env_root_path.parent)

        if not os.path.isdir(exe_env_root_path_parent):
            os.makedirs(exe_env_root_path_parent)

        exe_env_root_path_dir = exe_env_root_path.stem

        # Check if the directory exists and delete it (otherwise we can't clone into it)
        if os.path.isdir(exe_env_root_dir):
            logger.debug(f"Deleting existing configuration directory '{exe_env_root_dir}' to make way for cloning configuration repo")
            shell.run(exe_env_root_path_parent, rm, ["-rf", exe_env_root_path_dir])
        
        git_config = {"gitrepo":{"remote":config_git_repo, "default-branch":config_git_branch, "base-storage-dir":exe_env_root_path_parent}}
        logger.info(f"Cloning dynamic configuration from '{config_git_repo}' (on branch '{config_git_branch}')")
        try:

            perm_storage = GitPermanentStorage(git_config, execution_env)

            perm_storage.clone(exe_env_root_path_dir, config_git_branch)

            if delete_git_dir and exe_env_root_path.joinpath(".git").is_dir():
                shell.run(exe_env_root_dir, rm, ["-rf", ".git"])

            logger.info(f"Successfully cloned dynamic configuration from '{config_git_repo}' (on branch '{config_git_branch}')")

        except(StorageError):
            logger.error(f"Failed to download configuration from '{config_git_repo}' to '{exe_env_root_dir}'")
            # We tried and failed
            raise

        return

    @classmethod
    def getConfigPath(cls, config_path:str) -> str:
        """ Returns the config Path from the config base directory.  ConfigBase.init must have been called first. 

        If passed a path already relative to base_dir, just return it
        If passed an absolute path, 
          If relative to the install dir, append the relative path to the CWD to the base_dir.
          If relative to user home, and not an ephemeral env, just return it.
        If passed a relative path, it appends it directly to the base_dir
        """
        config_path_obj = Path(config_path).expanduser()
        if config_path_obj.absolute().is_relative_to(cls.base_dir):
            return str(config_path_obj.absolute())

        if config_path_obj.is_absolute():
            if config_path_obj.absolute().is_relative_to(cls.install_dir):

                return str(Path(cls.base_dir).joinpath(config_path_obj.relative_to(cls.install_dir)))

            elif not cls.ephemeral_env and config_path_obj.absolute().is_relative_to(Path.home()):

                # For persistent environments, configuration files should be able to exist in user's home directory
                return str(config_path_obj.absolute())

            else:
                return str(Path(cls.base_dir).joinpath(config_path_obj))
        else:
            return str(Path(cls.base_dir).joinpath(config_path_obj))


#!/usr/bin/env python3
"""
Validates whether the output of a string method matches an expected value
"""

import logging
from utils import match
from data.key import key as Key

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def validate(config:dict, key:Key, value:str, references:dict) -> bool:

    if (method := config.get("method", None)) is None:
        logger.error("Validator config did not have a 'method' key")
        return False
    
    if (expected_result := config.get("expected-result", None)) is None:
        logger.debug("Validator config did not have an 'expected-result' key. Result of method used directly.")

    parameter = config.get("input-parameter", None)
    if parameter is not None and not isinstance(parameter,str):
        logger.error("Validator config for 'input-parameter' must be a string (if specified)")
        return False

    if method == "equals":
        result = match.equals(value, expected_result)
    elif method == "notequals":
        result = not match.equals(value, expected_result)
    elif method == "isempty":
        result = match.is_empty(value)
    elif method == "isnotempty":
        result = not match.is_empty(value)
    elif method == "count":
        if parameter is None:
            logger.error("String method 'count' requires an 'input-parameter' to be configured")
            return False
        result = value.count(parameter) == expected_result
    elif method == "startswith":
        if parameter is None:
            logger.error("String method 'startswith' requires an 'input-parameter' to be configured")
            return False
        result = value.startswith(parameter)
    elif method == "endswith":
        if parameter is None:
            logger.error("String method 'endswith' requires an 'input-parameter' to be configured")
            return False
        result = value.endswith(parameter)
    elif method == "find":
        if parameter is None:
            logger.error("String method 'find' requires an 'input-parameter' to be configured")
            return False
        result = value.find(parameter) == expected_result
    elif method == "isalnum":
        result = value.isalnum()
    elif method == "isalpha":
        result = value.isalpha()
    elif method == "isascii":
        result = value.isascii()
    elif method == "isdecimal":
        result = value.isdecimal()
    elif method == "isdigit":
        result = value.isdigit()
    elif method == "islower":
        result = value.islower()
    elif method == "isnumeric":
        result = value.isnumeric()
    elif method == "isprintable":
        result = value.isprintable()
    elif method == "isspace":
        result = value.isspace()
    elif method == "istitle":
        result = value.istitle()
    elif method == "isupper":
        result = value.isupper()
    else:
        logger.error(f"Validator config 'method' key of '{method}' is not a supported string method")
        return False

    # if expected_result is not None:
    #     return_result = expected_result == result
    # else:
    #     return_result = result

    return result


#!/usr/bin/env python3
"""
Loads verifiers dynamically
"""

import logging
import importlib
from pathlib import Path
from utils.config import ConfigBase
from language.translate import Translate
from utils.load_yaml import yaml_file_to_dict

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class VerifiersConfig:
    """ Loads and store configuration options for Verifiers

    A class that uses yaml files to configure and discover verification methods.
    """

    VERIFIERS_DISPATCH_YAML = "verifiers_dispatch.yaml"
    VERIFIER_CONFIG_YAML = "verifiers_config.yaml"
    
    VERIFIERS_DISPATCH_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(VERIFIERS_DISPATCH_YAML))
    VERIFIER_CONFIG_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(VERIFIER_CONFIG_YAML))

    def __init__(self, verifiers_config:dict):
        #verifiers_dispatch_yaml_path = verifiers_config.get("verifiers_dispatch_yaml_config", self.VERIFIERS_DISPATCH_YAML_PATH)
        verifiers_dispatch_yaml_path = self.VERIFIERS_DISPATCH_YAML_PATH
        verifiers_config_yaml_path = ConfigBase.getConfigPath(verifiers_config.get("verifiers_config_yaml_config", self.VERIFIER_CONFIG_YAML_PATH))

        self.disable = verifiers_config.get("disable", [])
        if self.disable is None:
            self.disable = []

        self.validators_config = verifiers_config.get("validators", {})
        if self.validators_config is None:
            self.validators_config = {}

        self.dispatch = self._load_verifiers_dispatch(verifiers_dispatch_yaml_path)
        self.verifiers_config_dict = self._load_verifiers_config(verifiers_config_yaml_path)
        self.verifiers_texts_dict = self._load_verifiers_texts(ConfigBase.getConfigPath(self.verifiers_config_dict.get("output").get("template-text-file")))
        self.tag_mapping = self._load_tag_mapping(ConfigBase.getConfigPath(self.verifiers_config_dict.get("common").get("default-verifier-tag-mapping")))

    def _load_verifiers_config(self, verifiers_config_yaml_path) -> dict:
        
        yaml_config_dict = Translate.localiseYamlFile(verifiers_config_yaml_path)

        return yaml_config_dict["verifiers-config"]

    def _load_verifiers_dispatch(self, verifiers_dispatch_yaml_path:str):
        
        yaml_dict = yaml_file_to_dict(verifiers_dispatch_yaml_path) 
        
        all_verifiers = yaml_dict["verifiers-dispatch"]

        verifiers_dict = {}

        for verifier_name in all_verifiers:
            verifier_code = all_verifiers.get(verifier_name, "")
            if verifier_code == "":
                logger.warning(f"No value specified for verifier '{verifier_name}'")
                continue
            
            module_name = verifier_code
            if verifier_code.endswith(".py"):
                module_name = Path(__file__).parent.stem + "." + Path(verifier_code).stem
            logger.debug(f"Loading verifier file '{verifier_code}' as module '{module_name}'")
            # TODO catch ModuleNotFoundError exceptions
            imp = importlib.import_module(module_name)
            if hasattr(imp, "verify") and callable(imp.verify):
                verifiers_dict[verifier_name] = imp.verify
            else:
                logger.warning(f"Verifier file '{verifier_code}' did not have a 'verify' method")

        return verifiers_dict

    def _load_verifiers_texts(self, verifiers_texts_yaml_path:str):

        return yaml_file_to_dict(verifiers_texts_yaml_path) 

    def _load_tag_mapping(self, tag_mapping_yaml_path) -> list:
    
        #yaml_config_dict = yaml_file_to_dict(tag_mapping_yaml_path) 
        yaml_config_dict = Translate.localiseYamlFile(Path(tag_mapping_yaml_path)) 

        return yaml_config_dict["tag-mapping"]


#!/usr/bin/env python3
"""
Class MeasureOutput
"""

import logging
import utils.load_yaml
from utils.load_yaml import yaml_register_class
from utils.output import FormatOutput
import jsonpickle

import utils.logging
from utils.keymaster import get_column_name_for_key
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

# from jinja2 import Environment, FileSystemLoader, select_autoescape
# env = Environment(
#     loader = FileSystemLoader(searchpath="./"),
#     autoescape=select_autoescape()
# )

class Measurement:

    def __init__(self, config, this_model_title, this_model_version, other_model_title, other_model_version, section_key, tag_tuple, key_tag_list:list):
        yaml_register_class(Measurement)

        template_text_file = config.get("output").get("template-text-file")

        self.templated_texts = utils.load_yaml.yaml_file_to_dict(template_text_file).get("output-texts")
        self.config = config
        self.this_model_title = this_model_title
        self.this_model_version = this_model_version
        self.other_model_title = other_model_title
        self.other_model_version = other_model_version
        self.section_key = section_key
        self.tag_tuple = tag_tuple
        self.key_tag_list = key_tag_list
        self.count = 0
        self.distances = []

    def addCount(self, count):
        """ The total number of possible distances to measure for the tagged data """
        # Basically the number of rows
        self.count = count

    def addDistance(self, distance):
        self.distances.append(distance)

    def _get_state(self):

        output = {}
        
        context = {"section":self.section_key, "this_model_name":self.this_model_title, "other_model_name":self.other_model_title}
        measure_from_this_model = self.config["measure-tag"]["measure-this-model-against-other-model"]
        measure_from_other_model = self.config["measure-tag"]["measure-other-model-against-this-model"]
        measure_compare_missing = self.config["measure-tag"]["measure-compare-missing"]
        measure_compare_extra = self.config["measure-tag"]["measure-compare-extra"]

        measuring = {}
        measuring["section"] = self.section_key.getProperty("section")
        measuring["columns"] = []
        for col_name, tag_tuple in self.key_tag_list:
            column = {"name":get_column_name_for_key(col_name)}
            if tag_tuple[3] is not None:
                column["filter"] = tag_tuple[3]
            measuring["columns"].append(column)

        this_model = {"title":self.this_model_title, "version":self.this_model_version}
        other_model = {"title":self.other_model_title, "version":self.other_model_version}

        if measure_from_this_model in self.tag_tuple[1] and measure_compare_missing in self.tag_tuple[2]:
            output["from"] = this_model
            output["measuring"] = measuring
            output["to"] = other_model
        elif measure_from_this_model in self.tag_tuple[1] and measure_compare_extra in self.tag_tuple[2]:
            output["from"] = this_model
            output["measuring"] = measuring
            output["to"] = other_model
        elif measure_from_other_model in self.tag_tuple[1] and measure_compare_missing in self.tag_tuple[2]:
            output["from"] = other_model
            output["measuring"] = measuring
            output["to"] = this_model
        elif measure_from_other_model in self.tag_tuple[1] and measure_compare_extra in self.tag_tuple[2]:
            output["from"] = other_model
            output["measuring"] = measuring
            output["to"] = this_model

        output["distance-metric"] = len(self.distances)
        if len(self.distances) > 0:
            output["distances"] = []
            for row in self.distances:
                row_output = []
                for col_key, col_value in row:
                    col_output = {}
                    #col_output[get_column_name_for_key(col_key)] = col_value
                    col_output["column-name"] = get_column_name_for_key(col_key)
                    col_output["column-value"] = col_value
                    row_output.append(col_output)
                output["distances"].append(row_output)

        return output    

    def __getstate__(self):
        """ Used by jsonpickle to state of class to output """
        return self._get_state()

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node._get_state())

class MeasureOutput(FormatOutput):

    def __init__(self, output_config:dict):

        super().__init__(output_config)

        #template_text_file = output_config.get("template-text-file")

        #self.templated_texts = utils.load_yaml.yaml_file_to_dict(template_text_file).get("output-texts")

        self.measures = []

    def get_measure_metric(self) -> str:
        
        total_count = 0
        total_distances = 0
        for measurement in self.measures:
            total_count = total_count + measurement.count
            total_distances = total_distances + len(measurement.distances)
        return "{0:.0%}".format(total_distances/total_count)

    def getMeasureDetails(self):

        details = {}
        details["measure-metric"] = self.get_measure_metric()
        details["measurements"] = self.measures

        return details

    def _get_state(self):

        #details = {}
        #details["measure-metric"] = self.get_measure_metric()
        #details["measurements"] = self.measures
        self.details = self.getMeasureDetails()

        output = super()._get_state()

        return output

    def addMeasure(self, measurement:Measurement):
        self.measures.append(measurement)

    # def getSuccess(self, text_key:str, template_values:dict) -> dict:

    #     success_text = env.from_string(self.templated_texts.get(text_key)).render(template_values)

    #     return self._getOutput("Success", success_text, template_values["tm_version"])

    # def getError(self, text_key:str, template_values:dict) -> dict:

    #     error_text = env.from_string(self.templated_texts.get(text_key)).render(template_values)

    #     return self._getOutput("Error", error_text)

    def __getstate__(self):
        """ Used by jsonpickle to state of class to output """
        return self._get_state()

    def tojson(self):
        return jsonpickle.encode(self, unpicklable=False)


#!/usr/bin/env python3
"""
Verifies that all keys tagged as requiring value validation have valid values
"""

from data import find
import logging
from verifiers.verifier_error import ErrorType
from verifiers.verifier_error import VerifierIssue
from validators.validators import Validator
from validators.validator_output import ValidatorOutput
from utils import match

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


def verify(common_config:dict, verifier_config:dict, model:dict, template_model:dict) -> list:

    verify_return_list = []

    validator_config = verifier_config.get("validator-config", {})
    vlad = Validator(validator_config)

    non_mandatory_tag = verifier_config["not-mandatory-tag"]

    # Since a tagged key may have several validators configured (we OR the output), we don't find by tag,
    # but rather find by "validate-.*" glob, and then check each validator (as long as they are listed in config)

    # Need to get all keys for a given tag
    tagged_data = find.keys_with_tag_matching_regex(model, "^validate\\-.*$")

    configured_validators = verifier_config["tags"]
    logger.info(f"Configured validators in verifier config - {configured_validators}")

    references = {}
    references['model'] = model
    references['template-model'] = template_model

    for tagged_key, tagged_value in tagged_data:

        if match.is_empty(tagged_value) and tagged_key.hasTag(non_mandatory_tag):
            # Validations should be designed to fail with an empty value, but that's a problem if the value is not mandatory, 
            # so validations still succeed for empty values if the configured 'not-mandatory-tag' tag is present on the key
            continue

        tagged_value_validates = False
        a_validator_failed = False
        validatorOutput_list =[]

        # For each configured validator, check if it is a tag on the current key
        tagged_key_validator = [v for v in configured_validators if tagged_key.hasTag(v["tag"])]
        logger.debug(f"Validators configured for key '{tagged_key.name}' - {tagged_key_validator}")

        # Since we get data passed on tag prefix and not configured validators, it's possible that this tagged data has no configured validator.  No validator means no error
        at_least_one_validator_ran = False
        
        for validator_entry in tagged_key_validator:

            at_least_one_validator_ran = True

            validatorOutput = vlad.validate(validator_entry["tag"], tagged_key, tagged_value, references)

            # Culmultatively record success
            tagged_value_validates = tagged_value_validates or validatorOutput.validator_result

            # Track if a validator fails though
            if not validatorOutput.validator_result:
                a_validator_failed = True

            # Gather the responses from each validator so we can report them in the error
            validatorOutput_list.append(validatorOutput)
        
        issue_dict = {}
        issue_dict["issue_key"] = tagged_key
        issue_dict["issue_value"] = tagged_value
        issue_dict["errordata"] = validatorOutput_list

        if at_least_one_validator_ran:
            if not tagged_value_validates:    
                verify_return_list.append(VerifierIssue(
                    error_text_key="value-invalid", 
                    error_data_key="value-error-data",
                    fix_text_key=None, 
                    issue_dict=issue_dict))
            elif a_validator_failed:
                verify_return_list.append(VerifierIssue(
                    error_text_key="value-valid", 
                    error_data_key="value-error-data",
                    fix_text_key=None,
                    issue_dict=issue_dict,
                    errortype=ErrorType.INFO))

    return verify_return_list


#!/usr/bin/env python3
"""
Verifies that all keys tagged with references actually reference something
"""
import logging
from sre_parse import fix_flags
from utils import tags
from data.key import key as Key
from data import find
from verifiers.verifier_error import VerifierIssue
import verifiers.reference as reference
from language.translate import Translate
from utils import match, transform

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


def template_reference_callback(callback_config, tag_tuple, compare_value, compare_to_key, compare_to_value):

    #strip_config = {"start-char":"(", "end-char":")"}
    strip_config = callback_config["strip-context"]

    tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tag_tuple

    if tag_comparison == "template-approved":
        preApproved = compare_to_key.getProperty("templatePreApproved")
        # For pre-approved template values, we allow 'context' to be added in the TM e.g. in-memory (component).  But we strip this when matching to the pre-approved value
        #if preApproved is not None and match.equals(compare_value, compare_to_value, lambda val : _strip_context(strip_config, val)):
        if preApproved is not None and match.equals(compare_value, compare_to_value, transform.strip(strip_config["start-char"], strip_config["end-char"])):
            # TODO: Does not currently validate that the tag of the preApproved value matches the tag of the reference being checked e.g. pre-approved in functional assets table wouldn't match a ref tagged with just the the technical assets table.  This would restrict ref matches, so may not be a good thing.
            return True

    return False

def reference_callback(callback_config, tag_tuple, compare_value, compare_to_key, compare_to_value):

    tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tag_tuple

    strip_config = callback_config["strip-context"]
    strip_fn = transform.strip(strip_config["start-char"], strip_config["end-char"])

    # We want to strip any context for the purpose of copmarison of references
    if match.equals(compare_value, compare_to_value, strip_fn):
        return True

    if tag_comparison == "storage-expression":
        # TODO this just checks the value endswith the value from where it has a tag referencing, but really we should check for a complete
        # text match e.g. "All assets stored in environment variables"
        #grouped_text = callback_config.get("grouped-text", {}).get("storage-expression")

        # Since we are trying to match text we don't want the localised version to be output specific, we just want the default output format
        if match.starts_ends(compare_value, Translate.localise(callback_config.get("output-texts", {}), "start-assets-grouped-by-storage", ignore_format=True), strip_fn(compare_to_value)):
            return True
        if match.equals(compare_value, Translate.localise(callback_config.get("output-texts", {}), "all-assets", ignore_format=True)):
            return True

    # TODO- make this more generic so you can dynamically add more grouped text.
    if tag_comparison == "asset-expression":
        #asset_text = callback_config.get("grouped-text", {}).get("asset-expression")

        #if match.equals(compare_value, Translate.localise(callback_config.get("output-texts", {}), "all-functional-assets", cache_key="asset_text", ignore_format=True)):
        if match.equals(compare_value, Translate.localise(callback_config.get("output-texts", {}), "all-functional-assets", ignore_format=True)):
            return True

    return False

def _get_possible_fixed_texts(common_config:dict, key_entry:Key) -> list:

    # For each tag
    for tag in key_entry.getTags():

        tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tags.get_quad_tag_parts(tag)

        # TODO- make this more generic so you can dynamically add more grouped text.
        if tag_comparison == "asset-expression":

            #return Translate.localise(common_config["grouped-text"]["asset-expression"], "all-functional-assets")
            return Translate.localise(common_config["output-texts"], "all-functional-assets")

    return []

def verify(common_config:dict, verifier_config:dict, model:dict, template_model:dict) -> list:

    verify_return_list = []

    doc_reference_tag_prefix = common_config["references"]["doc-tag-prefix"]
    template_reference_tag_prefix = common_config["references"]["templ-tag-prefix"]

    reference_callback_config = {}
    #reference_callback_config["grouped-text"] = common_config["grouped-text"]
    reference_callback_config["output-texts"] = common_config["output-texts"]
    reference_callback_config["strip-context"] = common_config["strip-context"]

    # Find every key with a tag that starts with the configured prefix.
    all_reference_tagged_keys = find.keys_with_tag_matching_regex(model, "^(" + doc_reference_tag_prefix + "|" + template_reference_tag_prefix + ").*$")
           
    for key_entry, value_entry in all_reference_tagged_keys:

        # A quick sanity check that the value entry is a string
        if not isinstance(value_entry, str):
            logger.warning(f"A reference tag was applied to a field ('{key_entry.name}') whose value ('{value_entry}') is not a string. Ignoring.")
            continue
        
        # A key value might be allowed to reference several other fields e.g. asset might be functional or technical,
        # so success boolean will be true if any of the references is valid

        doc_reference_found = reference.check_reference(model, doc_reference_tag_prefix, key_entry, value_entry, reference_callback, reference_callback_config)
        doc_section_field_unref_error = []
        if not doc_reference_found:
            doc_section_field_unref_error = reference.get_reference_descriptions(model, doc_reference_tag_prefix, key_entry)

        template_reference_found = reference.check_reference(template_model, template_reference_tag_prefix, key_entry, value_entry, template_reference_callback, reference_callback_config)
        template_section_field_unref_error = []
        if not template_reference_found:
            template_section_field_unref_error = reference.get_reference_descriptions(template_model, template_reference_tag_prefix, key_entry)

        if not doc_reference_found and not template_reference_found:
            
            fixed_texts = _get_possible_fixed_texts(common_config, key_entry)

            issue_dict = {}
            issue_dict["issue_key"] = key_entry
            issue_dict["issue_value"] = value_entry
            issue_dict["fixdata"] = {}
            if len(doc_section_field_unref_error) > 0:
                issue_dict["fixdata"]["document"] = doc_section_field_unref_error
            if len(template_section_field_unref_error) > 0:
                issue_dict["fixdata"]["template"] = template_section_field_unref_error
            if len(fixed_texts) > 0:
                issue_dict["fixdata"]["static-texts"] = fixed_texts

            verify_return_list.append(VerifierIssue(
                error_text_key="reference-not-found", 
                fix_text_key="reference-not-found-fix", 
                fix_data_key="reference-not-found-fix-data", 
                issue_dict=issue_dict))

    return verify_return_list


#!/usr/bin/env python3
"""
Dynamically loads modules
"""

import logging
import importlib
from typing import Tuple
from utils.load_yaml import yaml_file_to_dict
from pathlib import Path

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


def load_from_dispatch_yaml(yamlpath:Path, rootkey:str, modulepath:Path, methodname:str) -> Tuple[dict, dict]:
    """ Dynamically loads a module and return a dict of methods"""

    yaml_dict = yaml_file_to_dict(yamlpath) 
        
    all_modules = yaml_dict[rootkey]
    
    methods_dict = {}
    modules_dict = {}

    for module_id, module_name in all_modules.items():

        fq_module_name = module_name
        if module_name.endswith(".py"):
            fq_module_name = modulepath.stem + "." + Path(module_name).stem
        logger.debug(f"Loading module file '{module_name}' as module '{fq_module_name}'")

        # TODO catch ModuleNotFoundError exceptions
        try:
            imp = importlib.import_module(fq_module_name)
        except ModuleNotFoundError as mnfe:
            logger.warning(f"Could not import module '{fq_module_name}'.  Error - {mnfe}")
            continue

        if hasattr(imp, methodname) and callable(getattr(imp, methodname)):
            methods_dict[module_id] = getattr(imp, methodname)
            modules_dict[module_id] = fq_module_name
        else:
            logger.warning(f"Validator module/file '{fq_module_name}' did not have a '{methodname}' method")
    
    return methods_dict, modules_dict


#!/usr/bin/env python3
"""
Abstracts away the execution provider e.g. local, AWS, GCP
"""

import logging
from pathlib import Path
from utils.config import ConfigBase
import utils.match as match
import utils.load_modules as load_modules
from utils.load_yaml import yaml_file_to_dict

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

PROVDERS_CONFIG_YAML = "providers_config.yaml"
PROVIDERS_CONFIG_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(PROVDERS_CONFIG_YAML))

PROVDERS_DISPATCH_YAML = "providers_dispatch.yaml"
PROVIDERS_DISPATCH_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(PROVDERS_DISPATCH_YAML))

def _load_config():

    yaml_config_dict = yaml_file_to_dict(ConfigBase.getConfigPath(PROVIDERS_CONFIG_YAML_PATH))

    return yaml_config_dict.get("providers", {})

def _load_dispatch():

    provider_loaders, _ = load_modules.load_from_dispatch_yaml(PROVIDERS_DISPATCH_YAML_PATH, "providers", Path(__file__).parent, "load")

    return provider_loaders

def get_provider(provider_name:str, no_config_mode:bool = False):

    provider_loaders = _load_dispatch()

    if not no_config_mode:
        provider_config = _load_config()
        return provider_loaders[provider_name](provider_config[provider_name])
    else:
        return provider_loaders[provider_name]({})






#!/usr/bin/env python3
"""
Classes to use git as a storage repository for threat model metadata
"""
import logging
from storage.gitrepo import GitStorage
import utils.match as match
from language.translate import Translate

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


class IndexStorage(GitStorage):

    def __init__(self, config:dict, execution_env, output_texts:dict = {}):
        super().__init__(config, execution_env)

        indexstorage_config = config.get("index", {})
        self.index_create_branch_name = indexstorage_config.get("index-create-branch", "create")
        self.output_texts = output_texts
        self.commit_message_key = indexstorage_config.get("index-commit-message-text-key")
        self.persist_changes = True

        self.entered = False

    def persistChanges(self, value:bool):
        self.persist_changes = value

    def __enter__(self):
        
        if self.entered:
            return self
        self.entered = True        

        super().__enter__()

        # Need to check if the remote branch has been merged i.e. git branch -r --merged approved, if so create a new branch
        # Once create branch merged the official approved branch may have been changed, so we need to use that, as opposed to the old create branch.  This
        # can still happen if the create branch has not been updated, but someone will be merging at some stage and can review
        if super().is_merged_to_default(self.index_create_branch_name):
            super().branch_replace(self.index_create_branch_name)
        else:
            super().branch_update(self.index_create_branch_name)

        return self

    def __exit__(self, exc_type, exc_value, traceback):

        if exc_type is None and self.persist_changes:
            if match.is_empty((commit_message := Translate.localise(self.output_texts, self.commit_message_key))):
                # Always have a default
                commit_message = f"Updated index file"
            super().commit(commit_message)

        super().__exit__(exc_type, exc_value, traceback)

        self.entered = False
    

class ThreatModelStorage(GitStorage):

    def __init__(self, config:dict, execution_env, ID:str, persist_changes:bool = True, output_texts:dict = {}):
        super().__init__(config, execution_env)

        tm_storage_config = config.get("threatmodel", {})
        self.ID = ID
        self.entered = False
        self.output_texts = output_texts
        self.commit_message_key = tm_storage_config.get("tm-commit-message-text-key")
        self.persist_changes = persist_changes

    
    def __enter__(self):

        if self.entered:
            return self
        self.entered = True

        super().__enter__()

        self.can_commit = False

        # Fetch the directory containing the metadata for the TM ID.  If it doesn't exist, nothing is downloaded
        super().checkout_directory(self.ID)

        # Overwrite any existing branch for making changes
        super().branch_replace(self.ID)

        return self

    def __exit__(self, exc_type, exc_value, traceback):

        if exc_type is None and self.persist_changes:
            if match.is_empty((commit_message := Translate.localise(self.output_texts, self.commit_message_key, {"ID":self.ID}))):
                # Always have a default
                commit_message = f"Update to {self.ID}"
            super().commit(commit_message)

        super().__exit__(exc_type, exc_value, traceback)

        self.entered = False

        return False


#!/usr/bin/env python3
"""
Verifies that all keys tagged DON'T have a, maybe specific, value, if some dependent key has a, maybe specific, value.  An opposite to conditional mandatory
"""

from data import find
import logging
from validators.validators import Validator
from verifiers.verifier_error import VerifierIssue
from utils import match
from utils import keymaster

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


def verify(common_config:dict, verifier_config:dict, model:dict, template_model:dict) -> list:

    validator_obj_config = common_config.get("validator-config", {})
    vlad = Validator(validator_obj_config)

    references = {}
    references['model'] = model
    references['template-model'] = template_model

    verify_return_list = []

    # The config could contain many tagged values to look for
    for conditional in verifier_config["conditionals"]:

        if_value_tag = conditional.get("if", {}).get("value-tag")
        then_value_tag = conditional.get("then", {}).get("value-tag")

        # Create an if_lambda parameterised on the 'if' Key and value
        if_value_condition = conditional.get("if", {}).get("value-condition")
        if_lambda = lambda key, value : vlad.validate_from_config(if_value_condition, key, value, references)

        # Create a then_lambda parameterised on the 'then' Key and value
        then_value_condition = conditional.get("then", {}).get("value-condition")
        then_lambda = lambda key, value : vlad.validate_from_config(then_value_condition, key, value, references)

        # Need to get all keys for a given 'if' tag
        if_tagged_data = find.keys_with_tag(model, if_value_tag)

        for if_tagged_key, if_tagged_value in if_tagged_data:
            # Get the rowID for the 'if' tagged key
            rowIDkey = keymaster.get_row_identifier_for_key(if_tagged_key)
            row_data = rowIDkey.getProperty("row")
            
            # Get the 'then' key and value
            then_tagged_key, then_tagged_value = find.key_with_tag(row_data, then_value_tag)

            # If the 'if' clause is true, but the 'then' clause is false, then that is a finding
            if if_lambda(if_tagged_key, if_tagged_value) and not then_lambda(then_tagged_key, then_tagged_value):

                issue_dict = {}
                issue_dict["issue_key"] = then_tagged_key
                issue_dict["issue_value"] = then_tagged_value
                if (issue_location := conditional.get("issue-location", None)) is not None:
                    # Override which of the 'if'/'then' keys is the 'issue_key', as sometimes errors can't report the 'then' key e.g. when testing for existence
                    if (issue_location == "if"):
                        issue_dict["issue_key"] = if_tagged_key
                        issue_dict["issue_value"] = if_tagged_value
                issue_dict["if_key"] = if_tagged_key
                issue_dict["if_value"] = if_tagged_value
                issue_dict["then_key"] = then_tagged_key
                issue_dict["then_value"] = then_tagged_value

                # The actual issue may be located somewhere else, so accomodate overrides to better explain how to fix the issue
                # Allow "issue-table-tag" to be set to allow override
                issue_table_value = model
                if (issue_table_tag := conditional.get("issue-table-tag", None)) is not None:
                    issue_table_key, issue_table_value = find.key_with_tag(model, issue_table_tag)
                    if issue_table_key is not None:
                        if (sectionKey := keymaster.get_section_for_key(issue_table_key)) is not None:
                            issue_dict["issue_table"] = sectionKey.getProperty("section")

                # Allow "issue-table-row" to be 'None' to indicate that no 'Location' is shown in the error
                if (issue_table_row := conditional.get("issue-table-row", 1)) is None:
                    if match.is_empty(issue_table_row):
                        issue_dict["issue_table_row"] = None

                # Use "issue-table-col-tag" to set the "issue_table_col" passed to VerifierIssue
                if (issue_key_tag := conditional.get("issue-table-col-tag", None)) is not None:
                    issue_key_key, _ = find.key_with_tag(issue_table_value, issue_key_tag)
                    if issue_key_key is not None:
                        issue_dict["issue_table_col"] = issue_key_key.getProperty("colname")
                        #issue_dict["issue_key"] = issue_key_key
                        #issue_dict["issue_value"] = None

                verify_return_list.append(VerifierIssue(
                    error_text_key=conditional["issue-text-key"], 
                        fix_text_key=conditional.get("fix-text-key", None), 
                        issue_dict=issue_dict))

    return verify_return_list


#!/usr/bin/env python3
"""
Verifies that all keys tagged as unique have a value that is unique
"""

from data import find
import logging
from data.key import key as Key
import utils.keymaster as keymaster
import utils.match as match
import verifiers.reference as reference
from verifiers.verifier_error import VerifierIssue

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


def reference_callback(callback_config, tag_tuple, compare_value, compare_to_key, compare_to_value):

    tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tag_tuple

    key_data_tag_name = callback_config.get("key_data_tag_name")

    if tag_comparison == "value_not_table":
        # Checking that tables don't match, but values do
        if key_data_tag_name != tag_data_tag_name and match.equals(compare_value, compare_to_value):
            return True

    return False

def verify(common_config:dict, verifier_config:dict, model:dict, template_model:dict) -> list:

    verify_return_list = []

    unique_tag_prefix = verifier_config["unique-tag-prefix"]

    previously_referenced = set()

    for key_entry, value_entry in find.keys_with_tag_matching_regex(model, "^" + unique_tag_prefix + ".*$"):

        # A quick sanity check that the value entry is a string
        if not isinstance(value_entry, str):
            logger.warning(f"A uniqueness tag was applied to a field ('{key_entry.name}') whose value ('{value_entry}') is not a string. Ignoring.")
            continue

        # Since non-unique values will always be in at least 2 places, we don't want to report errors for both places as really it's the same issue
        if key_entry in previously_referenced:
            continue

        callback_config = {"key_data_tag_name": keymaster.get_data_tag_for_key(key_entry)}

        referenced = reference.get_references(model, unique_tag_prefix, key_entry, value_entry, reference_callback, callback_config)

        # If len(referenced) > 1 then this key/value has been found in a different section
        # If len(referenced) == 1 then we do an object reference compare ("is") against what was found. reference.get_references will return only the first instance it finds of a matching key/value,
        # so the "referenced[0][1] is not key_entry" will be true when the current 'for' loop is the 2ND instance of the key/value combination i.e. 'for' loop is on duplicate, reference.get_references 
        # returns first instance, and these are different 'key' objects.  This is what finds duplicates in the same section/table
        if len(referenced) > 1 or (len(referenced) == 1 and referenced[0][1] is not key_entry):

            uniq_section_ref = reference.get_reference_descriptions(model, unique_tag_prefix, key_entry)

            referenced_data = [{'table':keymaster.get_section_for_key(key).getProperty("section"), 'column':key.getProperty("colname"), 'value':value} for tag, key, value in referenced if key is not key_entry]

            issue_dict = {}
            issue_dict["issue_key"] = key_entry
            issue_dict["issue_value"] = value_entry
            issue_dict["errordata"] = referenced_data
            issue_dict["fixdata"] = uniq_section_ref
            verify_return_list.append(VerifierIssue(
                error_text_key="not-unique",
                error_data_key="not-unique-data",
                fix_text_key="not-unique-fix", 
                fix_data_key="not-unique-fix-data",
                issue_dict=issue_dict))

            # Record already found duplicates
            previously_referenced.update([ref[1] for ref in referenced])

    return verify_return_list



#!/usr/bin/env python3
"""
Makes available configuration for CLI execution context
"""

import os
import configparser
import logging
from utils.error import ProviderError
from utils.config import ConfigBase

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class CLIContext:

    def __init__(self, config:dict):

        self.config = config
        
    def getConfluenceConnectionCredentials(self):
        
        confluence_config = configparser.ConfigParser()

        confluence_creds_file = ConfigBase.getConfigPath(self.config.get("confluence", {}).get("creds_config_file", ".atlassian"))
        confluence_creds_file = os.path.expanduser(confluence_creds_file)        
        confluence_config.read(confluence_creds_file)

        confluence_creds_section = self.config.get("confluence", {}).get("creds_config_section", "DEFAULT")

        confluence_config_dict = dict(confluence_config.items(confluence_creds_section))

        if len(confluence_config_dict) == 0:
            logger.error(f"The contents of the Confluence configuration file '{confluence_creds_file}' was empty")
            raise ProviderError("providers.emtpyConfluenceConfig", {"providers":{"creds_file":confluence_creds_file}})
    
        return confluence_config_dict

        
    def getGoogleCredentials(self):

        credentials = {}
        if (credentials_file := self.config.get("google", {}).get("credentials_file", None)) is None:
            # No credentials file specified
            if (api_key := self.config.get("google", {}).get("api_key", None)) is not None:
                credentials["api_key"] = api_key
            else:
                # Default to using the credentials.json file
                credentials_file = "credentials.json"
        
        if credentials_file is not None:
            credentials["credentials-file"] = ConfigBase.getConfigPath(credentials_file)
            credentials["token-file"] = ConfigBase.getConfigPath(self.config.get("google", {}).get("token_file", "token.json"))

            if not os.path.exists(credentials["credentials-file"]):
                logger.error(f"The Google credentials file '{credentials['credentials-file']}' does not exist")
                raise ProviderError("providers.noGoolgeCredsFile", {"providers":{"creds_file":credentials["credentials-file"]}})

        # For local testing with service credentials
        #svc_credentials_path = "convertors/gdoc_convertor/svc_credentials.json"
        #svc_credentials_fullpath = Path.joinpath(Path.cwd(), svc_credentials_path)
        #with open(svc_credentials_fullpath, 'r') as creds_file:
        #        credentials = json.load(creds_file)

        return credentials

    def getGitCredentials(self):

        # This should never be called for the CLI provider
        logger.error("Git credentials should not be requested for CLI provider")
        return None
        
        # For local container testing with git credentials
        # credentials = {}
        # with open("/root/.ssh/threatware_ed25519", 'r') as ssh_private_file:
        #     credentials["private-key"] = ssh_private_file.read()
        # with open("/root/.ssh/threatware_ed25519.pub", 'r') as ssh_public_file:
        #     credentials["public-key"] = ssh_public_file.read()

        # return credentials

    def setupGit(self, remote:str, base_storage_dir:str):

        # For the CLI we rely on the user's existing SSH credentials
        if remote.startswith("git@"):
            logger.info("Using user's existing git setup, including their SSH credentials for accessing git repositories")
        elif remote.startswith("http"):
            logger.info("Using git with anonymous HTTP")
        else:
            logger.warning("Unrecognised git remote, assuming anonymous HTTP")

    def get_config_base_dir(self, suggested_base_dir):
        """ Return the directory where we expect the configuration file to be based 
        
        Returns
        -------
        str : The suggested base directory to find configuration file
        bool : Whether this environment is ephemeral (config will not persist between invocations)
        """

        # Just return the suggestion whe using the CLI.  The config will persist after execution
        return suggested_base_dir, False


def load(config:dict):

    return CLIContext(config)


#!/usr/bin/env python3

import os
import logging

LOGGERNAME = 'threatware'

def getLoggerName(name):

    return LOGGERNAME + '.' + name

def configureLogging():

    # create logger
    logger = logging.getLogger(LOGGERNAME)
    
    #log_level = logging.DEBUG
    log_level = logging.WARNING

    if (env_log_level_str := os.getenv("THREATWARE_LOG_LEVEL")) is not None:
        env_log_level = logging.getLevelName(env_log_level_str)
        if isinstance(env_log_level, int):
            log_level = env_log_level
    
    logger.setLevel(log_level)

    logger.propagate = False

    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    # create formatter
    formatter = logging.Formatter('%(levelname)s:%(name)s:%(funcName)s:%(message)s')

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    logger.addHandler(ch)
    




# The pstr class extends the str class to allow str objects to have properties.
# The goal is to be able to use pstr objects instead of str objects when converting a threat model

def ensure_pstr(method):
    def wrapper(self, *args, **kwargs):
        result = method(self, *args, **kwargs)
        if isinstance(result, str) and not isinstance(result, pstr):
            result = pstr(result)
            if isinstance(self, pstr):
                result.properties = self.properties.copy()
        return result
    return wrapper

class pstr(str):

    _str_value = ""

    def __new__(cls, value, **kwargs):
        obj = super().__new__(cls, value)
        obj.properties = kwargs.get('properties', {})
        return obj

    def __init__(self, value, **kwargs):
        super().__init__()
        if type(value) == pstr:
            self._str_value = value._str_value
        else:
            self._str_value = value

    def to_str(self):
        # Return an instance of the superclass (str)
        return self._str_value
        #return super(pstr, self).__new__(str, self)

# List of str methods to decorate
str_methods = [
    'capitalize', 'casefold', 'center', 'encode', 'expandtabs', 'format', 'format_map',
    'join', 'ljust', 'lower', 'lstrip', 'maketrans', 'partition', 'replace', 'rfind',
    'rindex', 'rjust', 'rpartition', 'rsplit', 'rstrip', 'split', 'splitlines', 'strip',
    'swapcase', 'title', 'translate', 'upper', 'zfill', '__add__', '__contains__',
    '__eq__', '__ge__', '__getitem__', '__getnewargs__', '__gt__', '__hash__', '__iter__',
    '__le__', '__len__', '__lt__', '__mod__', '__mul__', '__ne__', '__repr__', '__rmod__',
    '__rmul__', '__str__'
]

# Decorate the methods
for method_name in str_methods:
    if hasattr(str, method_name):
        method = getattr(str, method_name)
        decorated_method = ensure_pstr(method)
        setattr(pstr, method_name, decorated_method)

# # Example usage
# s = pstr("hello", properties={"key1": "value1"})
# print(type(s.upper()))  # Output: <class '__main__.pstr'>
# print(s.upper().properties)  # Output: {'key1': 'value1'}
# print(type(s.replace("h", "j")))  # Output: <class '__main__.pstr'>
# print(s.replace("h", "j").properties)  # Output: {'key1': 'value1'}


#!/usr/bin/env python3

import logging
from data.data_config import DataConfig
from data.key import key as Key
from language.translate import Translate
import utils.match as match
import utils.model as model

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def parse(output_def):

    output_data_type = output_def.get("type", "list")
    if output_data_type == "dict":
        output = {}
    else:
        output = []

    return output

def assign(output_def, output_data):

    output_data_type = output_def.get("type", "list")
    if output_data_type == "dict":
        if isinstance(output_data, dict):
            return output_data
        if isinstance(output_data, list):
            if len(output_data) >= 1:
                if len(output_data) > 1:
                    logger.warning(f"dict output specified, but input was of length {len(output_data)} (>1). Just using first.")
                return output_data[0]

        logger.warning(f"Could not convert output data type '{type(output_data)}' to dict")
        return None
    else:
        return output_data

def _inherit_row_above_if_empty(proc_def, output_data):

    # output_data needs to be a list
    if not isinstance(output_data, list):
        logger.warning(f"Cannot run post-processor 'inherit-row-above-if-empty' on output data of type '{type(output_data)}', needs to be a list")
        return output_data

    # Get the key to look for
    if key := proc_def.get("key"):
        row_above_value = ""
        for row_item in output_data:

            if isinstance(row_item, dict):
                if key in row_item:
                    if isinstance(row_item[key], str):
                        value = row_item[key].strip()
                        if value == "":
                            # Replace value with value from row above
                            row_item[key] = row_above_value
                        else:
                            # We found a value for the current row, let's remember it for future rows
                            row_above_value = value

    return output_data

def _remove_empty_rows(proc_def, output_data):

    # output_data needs to be a list
    if not isinstance(output_data, list):
        logger.warning(f"Cannot run post-processor 'remove-empty-rows' on output data of type '{type(output_data)}', needs to be a list")
        return output_data

    # Get list of any columns to ignore the value of
    ignore_list = []
    if proc_def:
        ignore_list = proc_def.get("ignore-keys", [])

    def _inner_is_empty(key, value, context):
        if isinstance(value, str):
            if key.name not in ignore_list:
                if not match.is_empty(value):
                    return False, False

        return True, True

    new_output_data = []
    for row in output_data:
        empty = model.recurse(row, _inner_is_empty, None)
        # False will be returned if a non-empty found (we add these to the output), otherwise None will be returned if all empty
        if empty is False:
            new_output_data.append(row)

    return new_output_data

def _value_replace(proc_def, output_data):

    # output_data needs to be a str
    if not isinstance(output_data, str):
        logger.warning(f"Cannot run post-processor 'value_replace' on output data of type '{type(output_data)}', needs to be a string")
        return output_data

    # Get the key to look for
    for matches in proc_def:
        if match := matches.get("match"):
            if replacement := matches.get("replacement"):
                if output_data == match:
                    output_data = replacement
                    break   # exit early on first match

    return output_data

def _extract_col_name(colname:str) -> str:
    """
    Extracts all data before line boudnaries and just returns that as the column header.  This allows help text to be placed in the same table cell as a column header, but ignored.
    """
    if not isinstance(colname, str):
        return ""

    return colname.splitlines()[0]

def _remove_header_row(proc_def, output_data):
    """
    Removes the first row from the output data (must be a list).  Assumes values are column header names and sets the 'colname' property on all descendant 
    Keys.
    """

    # output_data needs to be a list
    if not isinstance(output_data, list):
        logger.warning(f"Cannot run post-processor 'remove-header-row' on output data of type '{type(output_data)}', needs to be a list")
        return output_data
    if len(output_data) == 0:
        logger.debug(f"Cannot run post-processor 'remove-header-row' on output data of no entries")
        return output_data

    # Get the list of col header names
    colnames = []
    for colname in output_data[0].values():
        # The value might not be a string, it might be a dictionary, but that dict should just have 1 entry
        if isinstance(colname, str):
            colnames.append(_extract_col_name(colname))
        else:
            if isinstance(colname, list) and len(colname) > 0:
                # Look at the first entry in the list, which should be a dict
                colname = colname[0]
            if isinstance(colname, dict):
                if len(colname.values()) == 0:
                    logger.warning(f"Expected 1 column name in dict '{colname}'.  Setting to empty")
                    colnames.append("")
                else:
                    # Table rows that aren't read in as straight values but instead are processed further, can have multiple values
                    #if len(colname.values()) > 1:
                    #    logger.warning(f"Expected only 1 column name in dict '{colname}'.  Using first")
                    colnames.append(_extract_col_name(list(colname.values())[0]))
            else:
                logger.warning(f"Unknown column name type of  '{type(colname)}'.  Setting to empty")
                colnames.append("")

    for row in output_data[1:]:
        for rowkey, rowvalue, colname in zip(row.keys(), row.values(), colnames):
            rowkey.addProperty("colname", colname)
            def _set_col_name(entry):

                if isinstance(entry, list):
                    for e in entry:
                        _set_col_name(e)
                if isinstance(entry, dict):
                    for k, v in entry.items():
                        if isinstance(k, Key):
                            k.addProperty("colname", colname)
                        _set_col_name(v)

            _set_col_name(rowvalue)

    return output_data[1:]

post_processor_dispatch_table = {
    "inherit-row-above-if-empty":_inherit_row_above_if_empty,
    "remove-empty-rows": _remove_empty_rows,
    "value-replace": _value_replace,
    "remove-header-row": _remove_header_row
}

def process(output_def, output_data):

    logger.debug(f'Entering: Keys {output_def.keys()}')

    # Run post-processing first
    if post_processor := output_def.get("post-processor"):
        logger.debug(f'Found post-processors {post_processor.keys()}')
        # Need to move through the post-processors in order specified in case order matters for processing
        for proc in post_processor.keys():
            if proc in post_processor_dispatch_table:
                output_data = post_processor_dispatch_table[proc](post_processor[proc], output_data)
            else:
                logger.warning(f"Could not find post-processor '{proc}'")

    # Mandatory output processing

    # Cleaning up 'n/a' lists
    # If the output value is a list with 1 entry, and that one entry is a dict with 1 key, and that key value is a string that matches 'n/a' -> then change to an empty list
    # This avoids the model looking like it contains a list entry of values, when really there is no value.
    # Need to mark changes like this, so verification knows that this change happened e.g. mandatory verification can ignore this.
    remove_na, apply_na_tags = DataConfig.remove_na_single_list_entry()
    if remove_na and isinstance(output_data, list):
        for list_entry in output_data:
            if isinstance(list_entry, dict):
                for entry_key, entry_value in list_entry.items():
                    if isinstance(entry_value, list) and \
                        len(entry_value) == 1 and \
                        isinstance(entry_value[0], dict) and \
                        len(entry_value[0]) == 1 and \
                        isinstance(list(entry_value[0].values())[0], str) and \
                        match.equals(list(entry_value[0].values())[0], Translate.getTranslation("na")):
                        list_entry[entry_key] = []
                        entry_key.addTags(apply_na_tags)


    # Change the output type last, as post-processing may have affected data structure
    if output_type := output_def.get("type"):
        output_data = assign(output_def, output_data)


    logger.debug(f'Leaving: Keys {output_def.keys()}')

    return output_data


#!/usr/bin/env python3
"""
Makes available configuration for API execution context
"""

import os
import configparser
from sh import ssh_keyscan
from sh.contrib import git
import utils.shell as shell
import logging
from utils.error import ProviderError
from utils.config import ConfigBase
from providers.cli import CLIContext

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

# API execution environment is the same as CLI, except it's an ephemeral environment
class APIContext:

    def __init__(self, config:dict):

        self.config = config
        self.clicontext = CLIContext(config)
        
    def getConfluenceConnectionCredentials(self):
        
        return self.clicontext.getConfluenceConnectionCredentials()

    def getGoogleCredentials(self):

        return self.clicontext.getGoogleCredentials()

    def getGitCredentials(self):

        # This should never be called for the API provider
        logger.error("Git credentials should not be requested for API provider")
        return None

    def setupGit(self, remote:str, base_storage_dir:str):

        # For the API we rely on the environment to have been already configured for git usage
        if remote.startswith("git@"):
            logger.info("Using local git setup, and local SSH credentials for accessing git repositories")
        elif remote.startswith("http:"):
            logger.info("Using git with anonymous HTTP")
        else:
            logger.warning("Unrecognised git remote, assuming anonymous HTTP")

        # Git must know who the user is before it can commit. Configure git user.name and user.email
        shell.run(os.path.expanduser("~/"), git.config, ["--global", "user.name", "threatware"])
        shell.run(os.path.expanduser("~/"), git.config, ["--global", "user.email", "threatware"])

    def get_config_base_dir(self, suggested_base_dir):
        """ Return the directory where we expect the configuration file to be based 
        
        Returns
        -------
        str : The suggested base directory to find configuration file
        bool : Whether this environment is ephemeral (config will not persist between invocations)
        """

        # Just return the suggestion when using the API.  The config will NOT persist after execution ('True' means ephemeral environment)
        return suggested_base_dir, True


def load(config:dict):

    return APIContext(config)


#!/usr/bin/env python3
"""
Reports on threats covering assets in location
"""

import logging
from utils import match
from utils.load_yaml import yaml_register_class
from verifiers.verifiers_config import VerifiersConfig
from data import find
from data import key as Key

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class CoveringThreat:

    def __init__(self, threat_description:str, threat_controls:list):
        yaml_register_class(CoveringThreat)

        self.threat_description = threat_description
        self.threat_controls = threat_controls

    def _get_state(self):

        output = {}
        output['threat_description'] = self.threat_description
        output['controls'] = []
        for control_key, control_value in self.threat_controls:
            output['controls'].append(control_value)
        
        return output

    def __getstate__(self):
        """ Used by jsonpickle to state of class to output """
        return self._get_state()

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node._get_state())

class CoveredAsset:

    def __init__(self, asset:Key, asset_value:str, storage_location:Key, storage_location_value:str, covering_threats:list):
        yaml_register_class(CoveredAsset)

        self.asset = asset
        self.asset_value = asset_value
        self.storage_location = storage_location
        self.storage_location_value = storage_location_value
        self.covering_threats = covering_threats

    def _get_state(self):

        output = {}
        output['asset'] = self.asset_value
        output['storage-location'] = self.storage_location_value
        output['covering-threats'] = self.covering_threats
        return output

    def __getstate__(self):
        """ Used by jsonpickle to state of class to output """
        return self._get_state()
        
    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node._get_state())


class ThreatCoverage:
    """ Searches a model for existing information regarding the threats covering assets in locations

    """

    def __init__(self, config:VerifiersConfig):
        self.config = config

    def coverage(self, model:dict) -> list:

        covered_assets = []

        common_config = self.config.verifiers_config_dict["common"]

        # Get a list of out-of-scope components
        out_of_scope_components = []
        components_key, components_data = find.key_with_tag(model, common_config["component-tags"]["component-data-tag"])
        for component_row in components_data:
            component_row_id_key, component_row_id_data = find.key_with_tag(component_row, "row-identifier")
            in_scope_key, in_scope_data = find.key_with_tag(component_row, common_config["component-tags"]["component-in-scope-tag"])
            if not match.equals(in_scope_data, common_config["component-tags"]["component-in-scope-value"]):
                out_of_scope_components.append(component_row_id_data)

        # Get all the data required to do the validation
        tagged_asset_data_list = common_config["asset-tags"]["asset-data-tag"]
        all_assets = []
        for tag in tagged_asset_data_list:
            tag_asset_data_key, tag_asset_data_value = find.key_with_tag(model, tag)
            if not isinstance(tag_asset_data_value, list):
                logger.error(f"Expecting the data tagged with '{tag}' to be a list, but it was a '{type(tag_asset_data_value)}'. Ignoring.")
                continue
        
            all_assets.append((tag, tag_asset_data_key, tag_asset_data_value))

        if len(all_assets) == 0:
            logger.error(f"Could not find any asset data tagged with '{common_config['asset-tags']['asset-data-tag']}'")

        # Loop through the asset data tables
        for asset_data_tag, asset_data_key, asset_data_value in all_assets:

            # Loop through the actual rows of assets in a data table
            for asset in asset_data_value:

                # Get the row identifier for the asset (to store covering threats against and for error reporting)
                row_id_key, row_id_data = find.key_with_tag(asset, "row-identifier")
                if row_id_key is None:
                    logger.error(f"Could not find a 'row-identifier' tag on any field for asset '{asset}'")

                # Get the storage-locations for the asset
                for storage_location_key, storage_location_value in find.keys_with_tag(asset, common_config["asset-tags"]['asset-location-tag']):

                    if storage_location_value in out_of_scope_components:
                        logger.debug(f"Ignoring coverage report on asset '{row_id_key.name}' in storage location '{storage_location_value}' as '{storage_location_value}' is out of scope")
                        continue

                    if storage_location_key.getProperty("excluded"):    # Returns None (evals to False) if not present, otherwise returns True/False
                        logger.debug(f"Ignoring coverage report on asset '{row_id_key.name}' in storage location '{storage_location_value}' it has a tag that excludes it")
                        continue

                    covering_threats = storage_location_key.getProperty("covering-threats")

                    covering_threats_list = []
                    if covering_threats is not None:
                        # Get required information about threat
                        for threat in covering_threats:

                            threat_description = find.key_with_tag(threat, common_config["threat-tags"]["threat-description-tag"])[1]
                            threat_controls = find.keys_with_tag(threat, common_config["threat-tags"]["threat-control-tag"])

                            covering_threats_list.append(CoveringThreat(threat_description, threat_controls))

                    covered_assets.append(CoveredAsset(row_id_key, row_id_data, storage_location_key, storage_location_value, covering_threats_list))

        return covered_assets


#!/usr/bin/env python3
"""
Validates whether a value is a date
"""

import logging
from pathlib import Path
import dateutil.parser 
from dateutil.parser import ParserError
from data.key import key as Key

import utils.logging
from validators.validator_output import ValidatorOutput
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def validate(config:dict, key:Key, value:str, references:dict) -> bool:

    try:
        parsed = dateutil.parser.parse(value)
        #key.addProperty(references['validator-tag'], config["output_text_valid"].format(key, value))
        return True
    except (ParserError, OverflowError):
        #key.addProperty(references['validator-tag'], config["output_text_invalid"].format(key, value))
        return False

    return False


#!/usr/bin/env python3
"""
Loads a scheme file
"""

import logging
from pathlib import Path
from utils.config import ConfigBase
from utils.error import SchemeError
from utils.load_yaml import yaml_file_to_dict

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

SCHEMES_YAML = "schemes.yaml"
SCHEMES_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(SCHEMES_YAML))

def load_scheme(template_scheme):

    yaml_dict = yaml_file_to_dict(ConfigBase.getConfigPath(SCHEMES_YAML_PATH)) 
    maps = yaml_dict["schemes"]
    
    logger.debug(f"Looking up scheme file for '{template_scheme}'")
    modelmap_file = maps.get(template_scheme, None)
    if modelmap_file is None:
        logger.error(f"Could not find scheme file for '{template_scheme}'")
        raise SchemeError("scheme.unknown-scheme", {"scheme":{"known_schemes":str(list(maps.keys())).replace("'", "")}})

    #yaml_dict = yaml_file_to_dict(str(Path(__file__).absolute().parent.joinpath(modelmap_file)))
    yaml_dict = yaml_file_to_dict(ConfigBase.getConfigPath(modelmap_file))
    return yaml_dict["scheme"]


# TODO write a validation routine for schemes.  
# For tags:
# - keys tags ending in "-data" should also have a section name
# - keys tags cannpt include a '/' (probably other characters as well) - anything that interferes with parsing, or might be a security issue e.g. {}
# - at least 1 field under a "-data" tagged section should have "row-identifier" set
# - tags of a form prefix/data-tag/entry-tag/method actually have data-tag, entry-tag values that are defined tags.  'method' must also exist.
# - currently must define "functional-asset-data", "technical-asset-data" and "threats-and-controls-data" for coverage validation to work
# For covertor:
# - check each processor defined has correct parameters for data.get, data.map and data.output e.g. 
# -- "inherit_row_above_if_empty" actually references an existing key
# -- "value-replace" match and replacement aren't both the same
# -- "value-extract" has a valid regex
# -- "column-num"/"row-num" are numeric


import logging
import re

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def key_with_tag(model, tag:str):

    if isinstance(model, dict):
        for dict_key, dict_value in model.items():
            if dict_key.hasTag(tag):
                return dict_key, dict_value
            if isinstance(dict_value, dict) or isinstance(dict_value, list):
                found_key, found_value = key_with_tag(dict_value, tag)
                if found_key is not None:
                    return found_key, found_value
    
    if isinstance(model, list):
        for list_entry in model:
            found_key, found_value = key_with_tag(list_entry, tag)
            if found_key is not None:
                return found_key, found_value

    return None, None

# Return a list of tuples (key,value) for all keys with the tag
def keys_with_tag(model, tag:str):

    keys_list = []

    if isinstance(model, dict):
        for dict_key, dict_value in model.items():
            if dict_key.hasTag(tag):
                keys_list.append((dict_key, dict_value))
            if isinstance(dict_value, dict) or isinstance(dict_value, list):
                keys_list.extend(keys_with_tag(dict_value, tag))
    
    if isinstance(model, list):
        for list_entry in model:
            keys_list.extend(keys_with_tag(list_entry, tag))

    return keys_list

# Return a list of tuples (key,value) for all keys with the tag
def keys_with_tag_matching_regex(model, regex_str:str, compiled_regex = None):
    """ 
    Returns a list of tuples (key,value) for all keys with a tag that matches the regular expression

    """

    if compiled_regex is None:
         compiled_regex = re.compile(regex_str)

    keys_list = []

    if isinstance(model, dict):
        for dict_key, dict_value in model.items():
            for tag in dict_key.getTags():
                if compiled_regex.search(tag) is not None:
                    keys_list.append((dict_key, dict_value))
                    break   # We only capture 1 tag.
            if isinstance(dict_value, dict) or isinstance(dict_value, list):
                keys_list.extend(keys_with_tag_matching_regex(dict_value, regex_str, compiled_regex))
    
    if isinstance(model, list):
        for list_entry in model:
            keys_list.extend(keys_with_tag_matching_regex(list_entry, regex_str, compiled_regex))

    return keys_list


#!/usr/bin/env python3
"""
Loads output format confgiuration
"""

import logging
from pathlib import Path
from utils.config import ConfigBase
import re
from language.translate import Translate

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class HTMLInject:

    def __init__(self, inject_config:dict):
            
        self.location_element = inject_config.get("location", {}).get("element", "")
        self.location_index = inject_config.get("location", {}).get("index", -1)  # By default assume -1, which means append to the end
        self.include_scripts = inject_config.get("include", {}).get("scripts", [])
        if self.include_scripts is None:
            self.include_scripts = []
        self.include_stylesheets = inject_config.get("include", {}).get("stylesheets", [])
        if self.include_stylesheets is None:
            self.include_stylesheets = []
        self.inline_element = inject_config.get("inline", {}).get("element", "")
        self.inline_style = inject_config.get("inline", {}).get("style", "")
        self.inline_script = inject_config.get("inline", {}).get("script", "")

class BannerConfig:

    def __init__(self, banner_config:dict):

        self.htmlinject = HTMLInject(banner_config.get("inject", {}))
        self.missingfindingstextkey = banner_config.get("text-keys", {}).get("missing-findings", "")
        
class FindingsConfig:

    def __init__(self, findings_config:dict):

        self.htmlinject = HTMLInject(findings_config.get("inject", {}))
        self.finding_attribute = findings_config.get("finding-attributes", {}).get("finding-index", "threatwarefinding")
        self.finding_type = findings_config.get("finding-attributes", {}).get("finding-type", "threatwarefinding-type")
        self.finding_class = findings_config.get("finding-attributes", {}).get("finding-class", "threatwarefinding-class")

class SchemeSpecificConfig:

    def __init__(self, scheme_specific:dict):

        self.htmlinject = HTMLInject(scheme_specific.get("inject", {}))
        self.regex = scheme_specific.get("regex", "")

    def matches(self, text:str):
        return re.search(self.regex, text) is not None

class HTMLConfig:

    def __init__(self, html_config:dict):

        self.schemeSpecificConfig = []
        for scheme_specific_entry in html_config.get("scheme-specific", []):
            self.schemeSpecificConfig.append(SchemeSpecificConfig(scheme_specific_entry))

        self.bannerConfig = BannerConfig(html_config.get("banner", {}))
        self.findingsConfig = FindingsConfig(html_config.get("findings", {}))

RESPONSE_CONFIG_YAML = "response_config.yaml"
    
RESPONSE_CONFIG_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(RESPONSE_CONFIG_YAML))

class ResponseConfig:
    """ Loads and store configuration options for Response

    A class that uses yaml files to configure the response.
    """
    def __init__(self):
        
        config = Translate.localiseYamlFile(ConfigBase.getConfigPath(RESPONSE_CONFIG_YAML_PATH))
        response_config = config.get("response-config", {})

        self.htmlConfig = HTMLConfig(response_config.get("html", {}))
        self.template_text_file = response_config.get("output").get("template-text-file")
        



#!/usr/bin/env python3
"""
Handler to invoke verifier
"""

import logging
import sys
import argparse
import configparser
from pathlib import Path
from importlib.metadata import version, PackageNotFoundError
from storage.gitrepo import GitStorage
from utils.error import ThreatwareError
from utils.request import Request
from utils.config import ConfigBase
from utils.output import FormatOutput
import utils.logging
from providers import provider
from language.translate import Translate
from schemes.schemes import load_scheme
import actions.convert as convert
import actions.verify as verify
import actions.manage as manage
import actions.measure as measure
from utils.output import FormatOutput, OutputType
from response.response import Response
from data.key import key as Key

utils.logging.configureLogging()
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

HANDLER_TEXTS_YAML = "handler_texts.yaml"
HANDLER_TEXTS_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(HANDLER_TEXTS_YAML))

ACTION_CONVERT = 'convert'
ACTION_VERIFY = 'verify'
ACTION_MANAGE_INDEXDATA = 'manage.indexdata'
ACTION_MANAGE_CREATE = 'manage.create'
ACTION_MANAGE_SUBMIT = 'manage.submit'
ACTION_MANAGE_CHECK = 'manage.check'
ACTION_MEASURE = 'measure'

def lambda_handler(event, context):

    # Fitler the set of query string parameters to just ones we know about
    if (qsp := event.get("queryStringParameters", {})) is None:
        qsp = {}
    Request.set(qsp)

    # Very first thing we need to do is find where all the configuration files are, and if they are not already present, download them.
    # How we do that depends what env we are in.  Providers usually take a config file, but we don't have them yet, so load without config (which limits what methods we can use)
    if getattr(context, "threatware.cli", False):
        execution_env = provider.get_provider("cli", no_config_mode=True)
    elif getattr(context, "threatware.api", False):
        execution_env = provider.get_provider("api", no_config_mode=True)
    else:
        # We are being called as a lambda, so get credentials from cloud
        execution_env = provider.get_provider("aws.lambda", no_config_mode=True)
    
    try:
        
        # This will set the ConfigBase.base_dir value so we can load config from this directory
        try:
            ConfigBase.init(execution_env)
        
        finally:
            # ConfigBase.init is guaranteed to configure at least the built-in config, which is enough to display at least an error.  Any exception in ConfigBase.init will
            # be captured by the outer try, but Translate/FormatOutput will be sufficiently configured in this finally: block to allow a properly formatted error message to
            # be returned to the user.

            ###
            # Now we can load things that have dynamic configuration
            ###

            # We need this to support localisation of keywords
            Translate.init()

            # Load the texts file with localised error messages
            handler_output = FormatOutput({"template-text-file":ConfigBase.getConfigPath(HANDLER_TEXTS_YAML_PATH)})
            # Create a response in case of an unexpected error
            response = Response(handler_output)
    

        # Validate input
        if Request.format == "html":
            if Request.action not in [ACTION_VERIFY]:
                Request.format = "json"
                logger.warning(f"'html' format is only supported for the 'verify' action, defaulting format to 'json'")
            if Request.reports != "none":
                Request.reports = "none"
                logger.warning(f"Additional reports are not supported in 'html' format, defaulting reports to 'none'")

        if Request.action is None:
            logger.error("action is a mandatory parameter")
            handler_output.setError("action-is-mandatory", {})
            response = Response(handler_output)
        elif Request.action not in [ACTION_CONVERT, ACTION_VERIFY, ACTION_MANAGE_INDEXDATA, ACTION_MANAGE_CREATE, ACTION_MANAGE_SUBMIT, ACTION_MANAGE_CHECK, ACTION_MEASURE]:
            logger.error(f"the action parameter must be one of {[ACTION_CONVERT, ACTION_VERIFY, ACTION_MANAGE_INDEXDATA, ACTION_MANAGE_CREATE, ACTION_MANAGE_SUBMIT, ACTION_MANAGE_CHECK, ACTION_MEASURE]}")
            handler_output.setError("action-value", {"actions":[ACTION_CONVERT, ACTION_VERIFY, ACTION_MANAGE_INDEXDATA, ACTION_MANAGE_CREATE, ACTION_MANAGE_SUBMIT, ACTION_MANAGE_CHECK, ACTION_MEASURE]})
            response = Response(handler_output)
        elif Request.action in [ACTION_CONVERT, ACTION_VERIFY, ACTION_MANAGE_CREATE, ACTION_MANAGE_SUBMIT, ACTION_MANAGE_CHECK, ACTION_MEASURE] and Request.scheme is None:
            logger.error("scheme is a mandatory parameter")
            handler_output.setError("scheme-is-mandatory", {})
            response = Response(handler_output)
        elif Request.action in [ACTION_CONVERT, ACTION_VERIFY, ACTION_MANAGE_CREATE, ACTION_MANAGE_SUBMIT, ACTION_MANAGE_CHECK, ACTION_MEASURE] and Request.docloc is None:
            logger.error("docloc is a mandatory parameter")
            handler_output.setError("docloc-is-mandatory", {})
            response = Response(handler_output)
        elif Request.action in [ACTION_VERIFY, ACTION_MEASURE] and  Request.doctemplate is None:
            logger.error(f"doctemplate is a mandatory parameter when action = {Request.action}")
            handler_output.setError("doctemplate-is-mandatory", {"action":Request.action})
            response = Response(handler_output)
        elif Request.action in [ACTION_MANAGE_INDEXDATA] and id is None:
            logger.error(f"ID is a mandatory parameter when action = {Request.action}")
            handler_output.setError("id-is-mandatory", {"action":Request.action})
            response = Response(handler_output)
        elif Request.action in [ACTION_MANAGE_CREATE] and Request.IDprefix is None:
            logger.error(f"IDprefix is a mandatory parameter when action = {Request.action}")
            handler_output.setError("idprefix-is-mandatory", {"action":Request.action})
            response = Response(handler_output)

        else:
            # Load the execution environment again, as this time we have configuration files to enable full functionality
            if getattr(context, "threatware.cli", False):
                # Get creds locally
                execution_env = provider.get_provider("cli")
            elif getattr(context, "threatware.api", False):
                execution_env = provider.get_provider("api")
            else:
                # We are being called as a lambda, so get credentials from cloud
                execution_env = provider.get_provider("aws.lambda")

            if Request.scheme is not None:
                schemeDict = load_scheme(Request.scheme)

            if Request.action == ACTION_CONVERT:
                
                convert_config = convert.config()

                # Convert the TM document
                convert_output = convert.convert(convert_config, execution_env, schemeDict, Request.docloc)
                response = Response(convert_output, force_api_format=True)     # In case convert failed

                if convert_output.getResult() != OutputType.ERROR:

                    doc_model = convert_output.getDetails()

                    verify_config = verify.config(schemeDict)

                    # Convert output is much more useful if it contains the default verify tags as well
                    # This will update the model in place
                    verify.assign_default_tags(verify_config, doc_model)

                response = Response(convert_output, Request.meta)

            elif Request.action == ACTION_VERIFY:

                convert_config = convert.config()
                
                # Convert the TM template
                convert_output = convert.convert_template(convert_config, execution_env, schemeDict, Request.doctemplate)
                response = Response(convert_output, force_api_format=True)     # In case convert failed

                if convert_output.getResult() != OutputType.ERROR:

                    template_model = convert_output.getDetails()

                    # Convert the TM document
                    convert_output = convert.convert(convert_config, execution_env, schemeDict, Request.docloc)
                    response = Response(convert_output, force_api_format=True)     # In case convert failed

                    if convert_output.getResult() != OutputType.ERROR:

                        doc_model = convert_output.getDetails()

                        verify_config = verify.config(schemeDict)

                        # Verify the TM document
                        verify_output = verify.verify(verify_config, doc_model, template_model)
                        response = Response(verify_output, force_api_format=True)     # In case verify failed

                        if verify_output.getResult() != OutputType.ERROR:
                            
                            issues = verify_output.getDetails()

                            # Generate a report on verification issues and analysis
                            verify_output = verify.report(verify_config, doc_model, issues, Request.reports)

                            response = Response(verify_output)

            elif Request.action == ACTION_MANAGE_INDEXDATA:
                
                manage_config = manage.config()

                output = manage.indexdata(manage_config, execution_env, Request.ID)

                response = Response(output)

            elif Request.action == ACTION_MANAGE_CREATE:
                
                convert_config = convert.config()

                # Convert the TM document
                convert_output = convert.convert(convert_config, execution_env, schemeDict, Request.docloc)
                response = Response(convert_output, force_api_format=True)     # In case convert failed

                if convert_output.getResult() != OutputType.ERROR:

                    doc_model = convert_output.getDetails()

                    manage_config = manage.config()

                    output = manage.create(manage_config, execution_env, Request.IDprefix, Request.scheme, Request.docloc, doc_model)

                    response = Response(output)

            elif Request.action == ACTION_MANAGE_CHECK:
                
                convert_config = convert.config()

                # Convert the TM document
                convert_output = convert.convert(convert_config, execution_env, schemeDict, Request.docloc)
                response = Response(convert_output, force_api_format=True)     # In case convert failed

                if convert_output.getResult() != OutputType.ERROR:

                    doc_model = convert_output.getDetails()
                
                    manage_config = manage.config()

                    # 'check' relies on measure
                    measure_config = measure.config()

                    output = manage.check(manage_config, execution_env, Request.docloc, Request.scheme, doc_model, measure_config, measure.distance)

                    response = Response(output)
            
            elif Request.action == ACTION_MANAGE_SUBMIT:
                
                convert_config = convert.config()

                # Convert the TM document
                convert_output = convert.convert(convert_config, execution_env, schemeDict, Request.docloc)
                response = Response(convert_output, force_api_format=True)     # In case convert failed

                if convert_output.getResult() != OutputType.ERROR:

                    doc_model = convert_output.getDetails()

                    manage_config = manage.config()

                    output = manage.submit(manage_config, execution_env, Request.docloc, Request.scheme, doc_model)

                    response = Response(output, "none")

            elif Request.action == ACTION_MEASURE:

                convert_config = convert.config()

                # Convert the TM template
                convert_output = convert.convert_template(convert_config, execution_env, schemeDict, Request.doctemplate)
                response = Response(convert_output, force_api_format=True)     # In case convert failed

                if convert_output.getResult() != OutputType.ERROR:

                    template_model = convert_output.getDetails()

                    # Convert the TM document
                    convert_output = convert.convert(convert_config, execution_env, schemeDict, Request.docloc)
                    response = Response(convert_output)     # In case convert failed

                    if convert_output.getResult() != OutputType.ERROR:

                        doc_model = convert_output.getDetails()

                        measure_config = measure.config()
                        
                        output = measure.distance_to_template(measure_config, execution_env, doc_model, template_model)

                        response = Response(output)
    
    except Exception as e:
        logger.error(e)
        if issubclass(type(e), ThreatwareError):
            handler_output.setError(e.text_key, e.template_values)
        else:
            handler_output.setError("internal-error", {})
        response = Response(handler_output, force_api_format=True)

    # Respond
    return {
        'statusCode': 200,
        "headers": {
            "Content-Type": f"{response.getContentType()}"
        },
        'body': response.getBody()
    }

def getVersion(prog):

    # This will search sys.path for threatware.egg-info/.dist-info directory to get the version from the PKG-INFO file
    try:
        
        version_number = version(prog)

        return version_number

    except PackageNotFoundError:
        pass

    # If the above fails then get the version from the local setup.cfg file
    setup_config_parser = configparser.ConfigParser()
    setup_config_parser.read(str(Path(__file__).absolute().parent.parent.joinpath("setup.cfg")))
    setup_config_dict = dict(setup_config_parser.items("metadata"))
    if "version" in setup_config_dict:
        return setup_config_dict["version"]

    return None

def main():

    scheme_help = 'Identifier for the threat model scheme (which contains location information)'
    doc_help = 'Location identifier of the document'
    template_help = 'Identifier for the document template (overrides template in scheme)'
    reports_help = "Additional reports can be returned with more information.\n'assets' will show the controls covering each asset per (in-scope) storage location.\n'controls' will show which assets each control covers per (in-scope) storage location"

    parser = argparse.ArgumentParser(prog='threatware', description='Threatware is a tool to help review threat models and provide a process to manage threat models.  It works directly with threat models as Confluence/Google Docs documents.  For detailed help on deployment, configuration and customisation, see https://threatware.readthedocs.io')

    version_str = f"{parser.prog} v{getVersion(parser.prog)}"
    parser.add_argument("-v", "--version", action="version", version=version_str)
    parser.add_argument("-l", "--lang", required=False, help="Language code for output texts")
    parser.add_argument("-f", "--format", required=False, help="Format for output, either JSON or YAML", default="json", choices=['json', 'yaml', 'html'])

    subparsers = parser.add_subparsers(dest="action", required=True)

    # convert
    parser_convert = subparsers.add_parser("convert", help='Convert a threat model for analysis')
    parser_convert.add_argument('-s', '--scheme', required=True, help=scheme_help)
    parser_convert.add_argument('-d', '--docloc', required=True, help=doc_help)
    parser_convert.add_argument("-m", "--meta", required=False, help="What level of meta data about fields should be returned.  Note, 'properties' returns 'tags' as well.", default="tags", choices=['none', 'tags', 'properties'])

    # verify
    parser_verify = subparsers.add_parser("verify", help='Verify a threat model is ready to be submitted for approval', formatter_class=argparse.RawTextHelpFormatter)
    parser_verify.add_argument('-s', '--scheme', required=True, help=scheme_help)
    parser_verify.add_argument('-d', '--docloc', required=True, help=doc_help)
    parser_verify.add_argument('-t', '--doctemplate', required=True, help=template_help)
    parser_verify.add_argument("-r", "--reports", required=False, help=reports_help, default='none', choices=['none', 'assets', 'controls', 'all'])

    # manage
    parser_manage = subparsers.add_parser("manage", help='Manage the status of threat models')
    manage_subparsers = parser_manage.add_subparsers(dest="manage_command", required=True)
    # manage.indexdata
    parser_manage_indexdata = manage_subparsers.add_parser("indexdata", help='Get the threat model index metadata for a threat model')
    parser_manage_indexdata.add_argument('-id', required=True, help='The document ID for the threat model index metadata to return')
    # manage.createdata
    parser_manage_create = manage_subparsers.add_parser("create", help='Create a new document ID for a new threat model')
    parser_manage_create.add_argument('-idprefix', required=True, help='The prefix of the document ID e.g. "CMP.TMD"')
    parser_manage_create.add_argument('-s', '--scheme', required=True, help=scheme_help)
    parser_manage_create.add_argument('-d', '--docloc', required=True, help=doc_help)
    # manage.check
    parser_manage_check = manage_subparsers.add_parser("check", help='Check whether the current threat model requires re-approval')
    parser_manage_check.add_argument('-s', '--scheme', required=True, help=scheme_help)
    parser_manage_check.add_argument('-d', '--docloc', required=True, help=doc_help)
    # manage.submit
    parser_manage_submit = manage_subparsers.add_parser("submit", help='Submit a threat model for approval')
    parser_manage_submit.add_argument('-s', '--scheme', required=True, help=scheme_help)
    parser_manage_submit.add_argument('-d', '--docloc', required=True, help=doc_help)
    
    # measure
    parser_measure = subparsers.add_parser("measure", help='Measure the distance of a TM from its template')
    parser_measure.add_argument('-s', '--scheme', required=True, help=scheme_help)
    parser_measure.add_argument('-d', '--docloc', required=True, help=doc_help)
    parser_measure.add_argument('-t', '--doctemplate', required=True, help=template_help)

    args = parser.parse_args()

    # Print out version string for the logs.  Do it after we parse the args so we don't print it twice if run with --version
    print(version_str, file=sys.stderr)

    # Build input for handler
    event ={}

    class Object(object):
        pass
    context = Object()
    setattr(context, "threatware.cli", True)

    action = args.action
    if action == "manage":
        action = action + "." + args.manage_command

    event["queryStringParameters"] = {}
    event["queryStringParameters"]["lang"] = args.lang if "lang" in args else None
    event["queryStringParameters"]["format"] = args.format if "format" in args else None
    event["queryStringParameters"]["action"] = action
    event["queryStringParameters"]["scheme"] = args.scheme if "scheme" in args else None
    event["queryStringParameters"]["docloc"] = args.docloc if "docloc" in args else None
    event["queryStringParameters"]["meta"] = args.meta if "meta" in args else None
    event["queryStringParameters"]["doctemplate"] = args.doctemplate if "doctemplate" in args else None
    event["queryStringParameters"]["ID"] = args.id if "id" in args else None
    event["queryStringParameters"]["IDprefix"] = args.idprefix if "idprefix" in args else None
    event["queryStringParameters"]["reports"] = args.reports if "reports" in args else None

    response = lambda_handler(event, context)

    print(response["body"])

    return 


if __name__ == "__main__":
    sys.exit(main())


#!/usr/bin/env python3

import logging
from lxml import etree
import lxml.html
from lxml.etree import XPathError, XPathEvalError
#from lxml.html.clean import clean_html
from html_table_parser import HTMLTableParser
#from html_table_extractor.extractor import Extractor
from utils import match
from utils.property_str import pstr
from html2text import HTML2Text

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

XPATH_FIELD = "xpath"
ATTR_NAME = "attribute-name"
COLUMN_NUM = "column-num"
ROW_NUM = "row-num"
PROCESS_IGNORE_COLS = "ignore-cols"
PROCESS_REMOVE_HEADER_ROW = 'remove-header-row'
PROCESS_REMOVE_ROWS_IF_EMPTY = "remove-rows-if-empty"
PROCESS_SPLIT_TYPE = "split-type"
HTML2TEXT_OPTIONS = "html2text-options"
HTML2TEXT_ALLOWLIST = [
    "unicode_snob",
    "escape_snob",
    "links_each_paragraph",
    "body_width",
    "skip_internal_links",
    "inline_links",
    "protect_links",
    "google_list_indent",
    "ignore_links",
    "ignore_mailto_links",
    "ignore_images",
    "images_as_html",
    "images_to_alt",
    "images_with_size",
    "ignore_emphasis",
    "bypass_tables",
    "ignore_tables",
    "google_doc",
    "ul_item_mark",
    "emphasis_mark", 
    "strong_mark",
    "single_line_break",
    "use_automatic_links",
    "hide_strikethrough",
    "mark_code",
    "wrap_list_items",
    "wrap_links",
    "wrap_tables",
    "pad_tables",
    "default_image_alt",
    "open_quote", 
    "close_quote", 
    "include_sup_sub"
]

def get_document(document_str, mapping):

    document = lxml.html.document_fromstring(document_str)

    if preprocess := mapping.get("preprocess", None):

        for processor in preprocess:
            selector = processor.get("selector", "//")

            try:
                # Navigate to the <table> element
                selection = document.xpath(selector)
            except XPathError:
                logger.warning(f"Pre-process selector '{selector}' caused an error")
                continue
            
            for ele in selection:
                # Order here matters, do attributes first because that is irresptive of tags or elements
                if attributes_to_strip := processor.get("strip-attributes", None):
                    etree.strip_attributes(ele, attributes_to_strip)
                # Do elements before tags because we may want to strip all tags and so if we did tags first we couldn't strip the elements
                if elements_to_strip := processor.get("strip-elements", None):
                    etree.strip_elements(ele, elements_to_strip)
                if tags_to_strip := processor.get("strip-tags", None):
                    etree.strip_tags(ele, tags_to_strip)
                
                

    # This is a hack.  For reasons unknown XPath queries start failing for no obvious reason
    # once strip_* methods are called.  So we reload the processed XML from a string.  Lame.
    new_doc_bstr = lxml.html.tostring(document, encoding='unicode')
    new_doc_str = new_doc_bstr
    document = lxml.html.document_fromstring(new_doc_str)

    return document


def get_document_section(document, query_cfg):

    if not query_key_defined(query_cfg, XPATH_FIELD):
        return None

    try:
        section_list = document.xpath(query_cfg[XPATH_FIELD])
    except XPathError:
        logger.warning(f"XPath query '{query_cfg[XPATH_FIELD]}' caused an error")
        section_list = []

    if len(section_list) == 0:
        logger.warning(f"XPath query '{query_cfg[XPATH_FIELD]}' returned no results")
        return None

    if len(section_list) != 1:
        logger.warning(f"XPath query '{query_cfg[XPATH_FIELD]}' returned more than 1 result, only using first")

    return section_list[0]

def get_document_list_items(document, query_cfg):

    if not query_key_defined(query_cfg, XPATH_FIELD):
        return []

    try:
        list_items = document.xpath(query_cfg[XPATH_FIELD])
    except XPathError:
        logger.warning(f"XPath query '{query_cfg[XPATH_FIELD]}' caused an error")
        list_items = []

    logger.debug(f"Query '{query_cfg[XPATH_FIELD]}' return {len(list_items)} elements")

    return list_items

def get_document_value(document, query_cfg) -> str:

    if not query_key_defined(query_cfg, XPATH_FIELD):
        return None

    try:
        value_list = document.xpath(query_cfg[XPATH_FIELD])
    except XPathError:
        logger.warning(f"XPath query '{query_cfg[XPATH_FIELD]}' caused an error")
        value_list = []

    logger.debug(f"Query '{query_cfg[XPATH_FIELD]}' return {len(value_list)} elements")

    if len(value_list) == 0:
        return None
    if len(value_list) > 1:
        logger.warning(f"Retreived {len(value_list)} document values when 1 was expected.  Using first.")
        
    location = ""

    if not isinstance(value_list[0], str):
        logger.warning(f"Expected results 'str', got '{type(value_list[0])}', for query '{query_cfg[XPATH_FIELD]}'")
    else:
        # It's a string, so we can't extract a location from it.  Try to strip away the xpath path that makes the result a string to get to the
        # element that contains the string.
        query = query_cfg[XPATH_FIELD]
        if query.endswith("//text()"):
            query = query[:-8]
        if query.endswith("/text()"):
            query = query[:-7]
        elif "/@" in query:
            query = query[0:query.rfind("/@")]
        
        try:
            value_element = document.xpath(query)
        except (XPathError, XPathEvalError):
            logger.warning(f"XPath query '{query}' caused an error")
            value_list = []
        if len(value_element) == 1:
            if hasattr(value_element[0], "xpath"):
                location = document.getroottree().getpath(value_element[0])
        
    # Despite checking this is a str, it could be a _ElementUnicodeResult, which doesn't serialise to JSON nicely.  So explicitly convert.

    if location:
        return pstr(str(value_list[0]), properties = {"location":location})
    
    logger.info(f"Returning value '{str(value_list[0])}' without location")
    return pstr(str(value_list[0]))

def get_element_text(document, query_cfg) -> str:

    if (element := _get_element(document, query_cfg)) is None:
        return None
    
    text = element.text_content()

    # The output from lxml is usually not a string, so explicitly convert
    return pstr(str(text), properties = {"location": element.getroottree().getpath(element)})


def get_element_attribute(document, query_cfg) -> str:

    if (element := _get_element(document, query_cfg)) is None:
        return None
    
    if not query_key_defined(query_cfg, ATTR_NAME):
        logger.warning(f"'{ATTR_NAME}' not specified in query configuration")

    attribute = element.get(query_cfg[ATTR_NAME], None)

    if attribute is None:
        logger.warning(f"Attribute '{query_cfg[ATTR_NAME]}' not found in element")

    # The output from lxml is usually not a string, so explicitly convert
    return pstr(str(attribute), properties = {"location": element.getroottree().getpath(element)})
    

def _remove_rows_if_empty(proc_def, table_data):

    # output_data needs to be a list
    if not isinstance(table_data, list):
        logger.warning(f"Cannot process 'remove_row_if_empty' on data of type '{type(table_data)}', needs to be a list")
        return table_data

    # Get list of any columns to ignore the value of
    ignore_list = []
    if proc_def:
        ignore_list = proc_def.get(PROCESS_IGNORE_COLS, [])

    new_table_data = []
    for row in table_data:
        test_row = row
        # If there are cols to ignore, then make a copy of the row without those cols
        if len(ignore_list) > 0:
            test_row = [value for count, value in enumerate(row) if count not in ignore_list]
        # Check if any of the values of the row are set
        if any(test_row):
            # At least one is set, so copy ORIGINAL row
            new_table_data.append(row)

    return new_table_data

def get_table_xpaths(table_element):
    row_xpaths = []
    roottree = table_element.getroottree()
    rows = table_element.xpath('.//tr')
    
    for row_index, row in enumerate(rows):
        col_xpaths = []
        cells = row.xpath('.//td | .//th')
        for cell_index, cell in enumerate(cells):
            # Since the cell element could be a td or th, we must make sure we get the correct xpath
            #cell_xpath = f"{table_element.getroottree().getpath(table_element)}//tr[{row_index + 1}]//{cell.tag}[{cell_index + 1}]"
            cell_xpath = roottree.getpath(cell)    # This is dangerous though as pre-processing may make the xpath here invalid when applied to the original document
            col_xpaths.append(cell_xpath)
        
        row_xpaths.append(col_xpaths)
    
    return row_xpaths

def get_document_row_table(document, query_cfg):

    if not query_key_defined(query_cfg, XPATH_FIELD):
        return None

    try:
        # Navigate to the <table> element
        table_list = document.xpath(query_cfg[XPATH_FIELD])
    except XPathError:
        logger.warning(f"XPath query '{query_cfg[XPATH_FIELD]}' caused an error")
        table_list = []

    if len(table_list) == 0:
        logger.warning(f"XPath query '{query_cfg[XPATH_FIELD]}' returned no results")
        return []
    elif len(table_list) != 1:
        logger.warning(f"XPath query '{query_cfg[XPATH_FIELD]}' returned more than 1 result, only using first")

    table_ele = table_list[0]

    # Get the XPath to each table element
    table_xpaths = get_table_xpaths(table_ele)

    # For future reference:
    #  'strip_tags' will remove the element tag and attributes but not the content
    #  'strip_elements' will remove everything, tag, attributes and content

    # Parsing a table is much easier if we remove HTML elements that just style the output, and since the output of
    # this method is no longer XML, we know subsequent queries don't rely on these elements.
    # Stripping span doesn't affect content (and Google Docs have A LOT of span elements)
    #etree.strip_tags(table_ele, "span")
    # Reluctantly removing 'superscript' elements.  It seems Google Docs will add single letter superscript to elements
    # when you download the doc as HTML, where that text is hihglight as part of a comment.  Lesser of two evils is to 
    # not support 'superscript' elements (as we have to allow a document with comments to be verified)
    #etree.strip_elements(table_ele, "sup")

    # When we call HTMLTableParser it will strip links i.e. <a> and replace them with data_separator, which is /n, 
    # this causes problems because if the link is in the middle of text it introduces a /n which messes up formatting.
    # We don't want to change data_separator because actual html breaks need to be preserved.  So we just strip links,
    # which is fine because we can't store them anyway and they should affect validation
    #etree.strip_tags(table_ele, "a")

    ## Removing this code as using HTMLTableParser with data_separator='\n' should have the same affecet
    # Stripping attributes of <p> elements doesn't affect content, but we leave them in because they are usually used
    # like newlines, which is what we replace them with further down.
    #for ele in table_ele.findall('.//p'):
    #for ele in table_ele.findall('.'):
    #    etree.strip_attributes(ele, "style")

    # Extract as string
    table_bstr = lxml.html.tostring(table_ele, encoding='unicode')

    ## TODO maybe use lxml.html.clean.clean_html in the future, but need to check how it handles imperfect HTML (I'm looking at you Confluence)
    #table_bstr = clean_html(table_bstr)
    
    ## Removing this code as using HTMLTableParser with data_separator='\n' should have the same affecet
    # HTMLTableParser will strip all tags from table entries, but sometimes that information is relevant e.g. <p></p> as newlines
    # This doesn't always work.  The '\n' need to be adjacent to actual text.  If the p-tags surround a <a> for instance then the '\n' is removed by HTMLTableParser
    #table_str = table_bstr.replace("</p><p>", "<a>\n</a>")
    #table_str = table_bstr.replace("</p><p>", "</p>\n<p>")
    table_str = table_bstr

    # Would like to use this libary, but it doesn't support changing the separator character.
    # extractor = Extractor(table_str)
    # extractor.parse()
    # table_output = extractor.return_list()

    # Non-breaking spaces cause string matching issues (usually from cut & paste from somewhere else).  So let's replace those with normal spaces.
    table_str = table_str.replace("\xa0", " ")

    # Parse as table
    # Tags in cells are stripped and the tags text content is joined using data_separator
    # HTMLTableParser defaults to decode_html_entities=False, but we want to keep those characters e.g. &
    p = HTMLTableParser(data_separator='\n', decode_html_entities=True)
    #p = HTMLTableParser(data_separator=' ')

    p.feed(table_str)

    if len(p.tables) != 1:
        logger.warning("Table parser returned more than 1 table, only using first")
    
    # Return list of lists.  There is only 1 table so only the first is returned
    table_output = p.tables[0]

    # Combine the XPath to each table cell wit the value in each table cell
    table_values = table_output
    if len(table_values) != len(table_xpaths):
        # We'll cope with this by using empty XPath vlaues
        logger.warning(f"Table values '{len(table_values)}' did not match the table xpaths '{len(table_xpaths)}'")
    for row_index, row in enumerate(table_values):
        if row_index >= len(table_xpaths):
            # Make entire row empty XPath values
            logger.warning(f"Table value row '{row_index}' did not have a corresponding table xpath")
            row_xpath = [""]*len(row)
        else:
            row_xpath = table_xpaths[row_index]
        for col_index, value in enumerate(row):
            if col_index >= len(row_xpath):
                # Make column XPath entry empty.  Remember, this row could contain real XPath values, just not enough of them
                logger.warning(f"Table value '{value}' did not have a corresponding table xpath")
                col_xpath = ""
            else:
                col_xpath = row_xpath[col_index]
            
            #table_output[row_index][col_index] = {"location":col_xpath, "value":value}
            table_output[row_index][col_index] = pstr(value, properties={"location":col_xpath})

    if query_cfg.get(PROCESS_REMOVE_HEADER_ROW, False):
        table_output = table_output[1:]

    # We need to detect is the key is present, but dict.get() will return 'None' if it's present but undefined.
    if PROCESS_REMOVE_ROWS_IF_EMPTY in query_cfg.keys():
        remove_rows_if_empty = query_cfg.get(PROCESS_REMOVE_ROWS_IF_EMPTY)
        table_output = _remove_rows_if_empty(remove_rows_if_empty, table_output)

    return table_output

def get_table_entry(row, query_cfg):

    if not query_key_defined(query_cfg, COLUMN_NUM):
        return None

    if query_cfg[COLUMN_NUM] >= len(row):
        logger.warning(f"Query configuration '{COLUMN_NUM}' with value '{query_cfg[COLUMN_NUM]}' was larger than the '{len(row)}' available columns (using zero-based indexing)")
        return None
    
    return row[query_cfg[COLUMN_NUM]]

# Split an existing table into multiple smaller tables
def get_split_table(table, query_cfg):

    output_tables = []

    if query_cfg.get(PROCESS_SPLIT_TYPE) == "distribute-header-col":
        header_col_num = query_cfg.get("header-col-num", 0)
        for row in table:
            for col_index, col_entry in enumerate(row[(header_col_num+1):]):
                if col_index >= len(output_tables):
                    output_tables.append([])
                output_table_entry = []
                output_table_entry.append(row[header_col_num])
                output_table_entry.append(col_entry)
                output_tables[col_index].append(output_table_entry)

    return output_tables

def get_transposed_table(table, query_cfg):

    return list(map(list, zip(*table)))

def get_remove_table_header_row(table, query_cfg):

    return table[1:]

# The query might constrain whether to return a value or not.  If not, None is returned.
def get_constrained_table_entry(row, query_cfg, current_row_index, current_col_index):
    
    if query_cfg is None:
        query_cfg = {}

    # If the constraint does not exist, then the default values ensure that the conditions are true.
    # If the constraint does exist, then it must be met.
    if query_cfg.get(ROW_NUM, current_row_index) == current_row_index:
           return row[query_cfg.get(COLUMN_NUM, current_col_index)]

    return None

def query_key_defined(query_cfg, query_cfg_key, optional = False):

    if query_cfg is None:
        logger.warning(f"Query configuration is not specified")
        return False

    if not isinstance(query_cfg, dict):
        logger.warning(f"Query configuration is not a set of key/value pairs")
        return False

    if query_cfg_key not in query_cfg:
        if not optional:
            logger.warning(f"Query configuration '{query_cfg_key}' was not present in '{query_cfg}'")
        return False
    
    return True

def _get_element(document, query_cfg) -> str:

    if not query_key_defined(query_cfg, XPATH_FIELD):
        return None

    try:
        value_list = document.xpath(query_cfg[XPATH_FIELD])
    except XPathError:
        logger.warning(f"XPath query '{query_cfg[XPATH_FIELD]}' caused an error")
        value_list = []

    logger.debug(f"Query '{query_cfg[XPATH_FIELD]}' return {len(value_list)} elements")

    if len(value_list) == 0:
        return None
    if len(value_list) > 1:
        logger.warning(f"Retreived {len(value_list)} document values when 1 was expected.  Using first.")

    return value_list[0]

def does_col_index_match(query_cfg, index):

    if not query_key_defined(query_cfg, COLUMN_NUM):
        return False

    cfg_index = query_cfg[COLUMN_NUM]
    if cfg_index:
        # There is a value to match
        return cfg_index == index
    
    # There is no value, so no match
    return False

def get_row_entry_by_col_index(row, index):

    if index >= len(row):
        logger.warning(f"Index value '{index}' was larger than the '{len(row)}' available rows (using zero-based indexing)")
        return None

    return row[index]

def set_table_entry(row, query_cfg, value):

    if not query_key_defined(query_cfg, COLUMN_NUM):
        return

    row[query_cfg[COLUMN_NUM]] = value

    return

# Expecting the xpath to return a list of nodes, which get text exrtacted and concatenated
def get_text_section(document, query_cfg):

    if not query_key_defined(query_cfg, XPATH_FIELD):
        return None

    try: 
        nodes = document.xpath(query_cfg[XPATH_FIELD])
    except XPathError:
        logger.warning(f"XPath query '{query_cfg[XPATH_FIELD]}' caused an error")
        nodes = []

    input = ""
    for node in nodes:
        input += lxml.html.tostring(node, encoding='unicode')

    h2t = HTML2Text()
    if query_key_defined(query_cfg, HTML2TEXT_OPTIONS, optional=True) and isinstance(query_cfg[HTML2TEXT_OPTIONS], dict):
        for option, value in query_cfg[HTML2TEXT_OPTIONS].items():
            if isinstance(option, str) and option.lower() in HTML2TEXT_ALLOWLIST:
                setattr(h2t, option.lower(), value)
            else:
                logger.warning(f"html2text option '{option.lower()}' not in allowed list")
    
    output = h2t.handle(str(input))

    return pstr(output, properties={"location":query_cfg[XPATH_FIELD]})

html_dispatch_table = {
    "html-ul":get_document_list_items,
    "html-table":get_document_row_table,
    "html-table-remove-header-row":get_remove_table_header_row,
    "html-table-row":get_table_entry,
    "html-split-table":get_split_table,
    "html-table-transpose":get_transposed_table,
    "html-text":get_document_value,
    "html-element-text":get_element_text,
    "html-element-attribute": get_element_attribute,
    "html-text-section":get_text_section
}


#!/usr/bin/env python3
"""
Utility methods for string matching
"""

import logging
from utils import transform
from utils.model import recurse

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

# def c14n(value:str, transform_fn = None) -> str:
#     if transform_fn is not None:
#         value = transform_fn(value)
#     return value.casefold().strip()

# Returns the 'possible' value if 'str_to_match' equals 'possible' string/one of list of strings
def get_equals(str_to_match:str, possible, transform_fn = None) -> str:

    if str_to_match is None:
        str_to_match = ""
    if possible is None:
        possible = ""

    if isinstance(str_to_match, str):
        str_to_match_c14n = transform.c14n(str_to_match, transform_fn)
    else:
        logger.warning(f"Expecting a string to match but '{str_to_match}' is a '{type(str_to_match)}'")

    if isinstance(possible, str):
        possible = [possible]

    for possible_match in possible:
        if possible_match is None:
            possible_match = ""
        if isinstance(possible_match, str):
            possible_match_c14n = transform.c14n(possible_match, transform_fn)
        if str_to_match_c14n == possible_match_c14n:
            return possible_match

    return None

# Returns True if 'str_to_match' equals 'possible' string/one of list of strings
def equals(str_to_match:str, possible, transform_fn = None) -> bool:
    return get_equals(str_to_match, possible, transform_fn) is not None

# Returns True if 'str_to_match' is empty
def is_empty(str_to_match:str) -> bool:

    return equals(str_to_match, "")

# Returns True if all descendents of 'data' are empty
def is_empty_recursive(data:object) -> bool:

    def _inner_is_empty(key, value, context):
        if isinstance(value, str):
            if is_empty(value):
                return True, True
            else:
                return False, False

    _, result = recurse(data, _inner_is_empty, None)

    return result

def endswith(str_to_match:str, ends_with) -> bool:

    str_to_match = transform.c14n(str_to_match)

    if isinstance(ends_with, str):
        ends_with = [ends_with]
    if not isinstance(ends_with, list):
        logger.error(f"'ends_with' parameter must be string or list of strings, not '{type(ends_with)}'")

    does_end_with = False
    for ending_str in ends_with:
        if str_to_match.endswith(transform.c14n(ending_str)):
            does_end_with = True
            break

    return does_end_with
    

# Returns True if string starts with 'start_with' string/one of list of strings, and ends with 'ends_with' string/one of list of strings
def starts_ends(str_to_match:str, starts_with, ends_with) -> bool:

    str_to_match = transform.c14n(str_to_match)

    if isinstance(starts_with, str):
        starts_with = [starts_with]
    if isinstance(ends_with, str):
        ends_with = [ends_with]
    if not isinstance(starts_with, list):
        logger.error(f"'starts_with' parameter must be string or list of strings, not '{type(starts_with)}'")
    if not isinstance(ends_with, list):
        logger.error(f"'ends_with' parameter must be string or list of strings, not '{type(ends_with)}'")

    does_start_with = False
    for starting_str in starts_with:
        if str_to_match.startswith(transform.c14n(starting_str)):
            does_start_with = True
            break
    
    does_end_with = False
    for ending_str in ends_with:
        if str_to_match.endswith(transform.c14n(ending_str)):
            does_end_with = True
            break

    return does_start_with and does_end_with

def get_contains(str_to_match:str, values, transform_fn = None) -> bool:
    """
    Returns the value, if any of the values are contained in str_to_match
    """

    str_to_match = transform.c14n(str_to_match, transform_fn)

    if isinstance(values, str):
        values = [values]
    if not isinstance(values, list):
        logger.error(f"'values' parameter must be string or list of strings, not '{type(values)}'")
    
    for value_str in values:
        if str_to_match.find(transform.c14n(value_str, transform_fn)) != -1:
            return value_str
            
    return None

def contains(str_to_match:str, values, transform_fn = None) -> bool:
    """
    Returns True, if any of the values are contained in str_to_match
    """
    return get_contains(str_to_match, values, transform_fn) is not None



#!/usr/bin/env python3

import logging
from data.data_config import DataConfig
import data.value
import convertors.html_convertor.query as query

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def doc_to_model(config:dict, document:dict, mapping:dict):
    
    tm = {}

    try:
        
        DataConfig.init(config.get("data", {}))

        tm = data.value.parse(mapping['map'], document)

    except BaseException as err:
        logger.error(f"Unexpected err='{err}', type='{type(err)}'")
        raise

    return tm



#!/usr/bin/env python3

import logging
import data.get as get
import data.output
import data.map

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def parse(value_def, input):

    logger.debug(f'Entering: value_def = {value_def.keys() if value_def else None}')

    if not value_def:
        # There is nothing defined to change the input, so we can just return the input
        return input

    output = None

    #   Process scheme field 
    #       get-data: 
    #           query ...
    #

    # If there is a get-data section use that to extract the data we want to map
    data_value = get.parse(value_def.get("get-data",{}), input)

    #   Process scheme field 
    #       map-data: 
    #           - key ...
    #             value ...
    #

    # If there is a map-data section we'll use that to map data to output
    map_data_def = value_def.get("map-data", [])
    
    # If there is no mapping, then we just return the value (no more recursion into this method)
    if map_data_def is None or len(map_data_def) == 0:
        logger.debug("No map_data so directly passing output of get-data to output-data")
        output = data_value
    else:
        # The data returned by get-data needs to be mapped to correct values.  If get-data returned a list then each item in the list
        # needs to be passed for mapping
        if isinstance(data_value, list):
            logger.debug(f"A list of {len(data_value)} data inputs will be mapped")
            
            # If the data_value is a list, then for each item in the list we will apply the map_data_def.  We track output in a list.
            output_list = []
            for list_item in data_value:

                # Map an item of the data list
                # If map-data:value exists it will call back into this method with the list item as input
                output_dict = data.map.parse(map_data_def, list_item)
                
                logger.debug(f"Mapped data {output_dict}")
                output_list.append(output_dict)
            
            output = output_list
        else:
            
            # Map the data
            output_dict = data.map.parse(map_data_def, data_value)

            logger.debug(f"Mapped data {output_dict}")
            
            output = output_dict

    #   Process scheme field 
    #       output-data: 
    #           type: ...
    #           post-processor: ...
    #

    output_data_def = value_def.get("output-data", {})
    output = data.output.process(output_data_def, output)

    logger.debug(f'Leaving: value_def = {value_def.keys()}')

    return output


#!/usr/bin/env python3
"""
Classes to manipulate for threat model metadata
"""
import logging
from pathlib import Path
from data.key import KeySerialiseType, key as Key
from utils import keymaster, match, tags, load_yaml
from utils.load_yaml import yaml_register_class
from data import find
from utils.error import ManageError
from manage.manage_storage import ThreatModelStorage

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class MetadataIndexEntry:
    """
    This class stores the index metadata associated to a version of the threat model.  
    
    Its purpose is to hold sufficient data to determine if a threat model is Approved (and to identify that threat model). 
    """

    def __init__(self, config:dict, entry:dict):
        yaml_register_class(MetadataIndexEntry)

        self.approvedStatus = config.get("approved-status", "Approved")
        self.draftStatus = config.get("draft-status", "Draft")

        self.ID = entry.get("ID", None)
        self.title = entry.get("title", None)
        self.scheme = entry.get("scheme", None)
        self.location = entry.get("location", None)
        self.approved_version = entry.get("approved-version", None)
        self.version_status = entry.get("version-status", None)
        self.approver = entry.get("approver", None)
        self.approved_date = entry.get("approved-date", None)

    def _populate(self, scheme, location, model, version_tag):
        """ 
        This populates using the current version data from the model 
        
        """

        _, ddd_value = find.key_with_tag(model, "document-details-data")
        _, self.title = find.key_with_tag(ddd_value, "document-title")
        _, self.ID = find.key_with_tag(ddd_value, "document-id")
        _, version_tag_value = find.key_with_tag(ddd_value, version_tag)

        self.scheme = scheme
        self.location = location

        _, vhd_value = find.key_with_tag(model, "version-history-data")
        versions = find.keys_with_tag(vhd_value, "row-identifier")
        
        # If the version value is empty then there will be no match in the version history table
        if not match.is_empty(version_tag_value):
            for version_key, version_value in versions:
                if match.equals(version_tag_value, version_value):
                    version_row = version_key.getProperty("row")
                    self.version_status = find.key_with_tag(version_row, "version-status")[1]
                    self.approver = find.key_with_tag(version_row, "version-approver")[1]
                    self.approved_date = find.key_with_tag(version_row, "version-approved-date")[1]
                    break

            if self.isApproved():
                self.approved_version = version_tag_value

        return version_tag_value

    def createCurrentVersionEntryFromModel(self, scheme, location, model):
        """ 
        This populates using the current version data from the model 
        
        """

        return self._populate(scheme, location, model, "current-version")

    def createApprovedVersionEntryFromModel(self, scheme, location, model):
        """ 
        This populates using the approved version data from the model 
        
        """

        return self._populate(scheme, location, model, "approved-version")

    def isApproved(self) -> bool:
        """ An entry is considered approved if:
             - status is (localised) 'Approved'
             - approver is set
             - approved-date is set
        """
        if match.equals(self.version_status, self.approvedStatus)  and \
           self.approver is not None                       and \
           self.approved_date is not None:
            return True

        return False

    def hasApproval(self) ->  bool:
        """ Return true if an approver name and approved date exists for the threat model """
        if self.approver is not None and self.approved_date is not None:
            return True
        else:
            return False

    def stripApproval(self):

        self.approved_version = ""
        self.version_status = self.draftStatus
        self.approver = ""
        self.approved_date = ""

    def versionStatus(self) -> str:
        return self.version_status

    def get_state(self):
        return self._get_state()

    def _get_state(self):
        output = {}
        output["ID"] = self.ID
        output["title"] = self.title
        output["scheme"] = self.scheme
        output["location"] = self.location
        output["approved-version"] = self.approved_version
        output["version-status"] = self.version_status
        output["approver"] = self.approver
        output["approved-date"] = self.approved_date

        return output

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node._get_state())   


class ThreatModelVersionMetaData:
    """
    This class stores the metadata associated to a version of the threat model.  
    
    It could be any data in the threat model (tags are used to determine what data) and isn't necessarily sufficient data to determine if a threat model is Approved. 
    """
    def __init__(self, config = {}) -> None:
        yaml_register_class(ThreatModelVersionMetaData)

        self.version_meta_tag_prefix = config.get("version-meta-tag-prefix", "version-meta")
        self.version_meta = {}
        
        self.version = None
        
    def fromModel(self, scheme, location, model):

        # Find all the keys prefixed with tag "version-meta"
        for value_key, value_data in find.keys_with_tag_matching_regex(model, "^" + self.version_meta_tag_prefix):
            if value_data is not None:
                # Get the full tag that matches
                for tag in value_key.getTags():
                    if match.starts_ends(tag, self.version_meta_tag_prefix, ""):
                        if match.equals(tag, self.version_meta_tag_prefix):
                            self.version_meta[value_key.name] = value_data
                        else:
                            # meta data will only include a subset of the tagged data
                            if isinstance(value_data, list):
                                tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tags.get_quad_tag_parts(tag)
                                data_model_key, data_model_value = find.key_with_tag(model, tag_data_tag_name)
                                data_key, data_value = find.key_with_tag(data_model_value, tag_field_tag_name)
                                if len(value_data) > 0:
                                    row_identifier = keymaster.get_row_identifier_for_key(list(value_data[0].keys())[0])
                                    for row_dict in value_data:
                                        if match.equals(row_dict[row_identifier], data_value):
                                            self.version = data_value
                                            self.version_meta = self.version_meta | row_dict

        self.version_meta = self.version_meta | {"scheme":scheme, "location":location}

    def fromDict(self, version:str, meta_dict:dict) -> None:

        self.version = version
        self.version_meta = meta_dict

    def _get_state(self):
        return self.version_meta

    def __getstate__(self):
        """ Used by jsonpickle to state of class to output """
        return self._get_state()

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node._get_state())


class IndexMetaData:
    """
    Stores all MetadataIndexEntry for all versions of a threat model.

    Can retrieve entries, update new entries and persist all entries
    """

    def __init__(self, config:dict, storage) -> None:
        yaml_register_class(IndexMetaData)

        self.config = config
        self.index_filename = config.get("index-filename", "threatmodels.yaml")
        self.storage = storage

        self._load_index()

    def _load_index(self):

        self.indexdata = {}

        if (threatmodels := self.storage.load_yaml([], Path(self.storage.repodir).joinpath(self.index_filename))) is not None:
            for TMindex_list_entry in threatmodels.get("threatmodels", []):
                mdie = MetadataIndexEntry(self.config, TMindex_list_entry)
                self.indexdata[mdie.ID] = mdie
        else:
            logger.warning(f"No {self.index_filename} was loaded.  This should only happen when the metadata repo is first created and is empty.")

    def getIndexEntry(self, ID:str) -> MetadataIndexEntry:
        return self.indexdata.get(ID, None)

    def getIndexEntryByLocation(self, location:str) -> MetadataIndexEntry:

        location_based_index = {entry.location:entry for entry in self.indexdata.values()}

        return location_based_index.get(location, None)

    def getIDByLocation(self, location:str) -> str:

        location_based_index = {entry.location:ID for ID, entry in self.indexdata.items()}

        return location_based_index.get(location, None)

    def createIndexEntryID(self, IDprefix:str):

        #last_entry_ID = list(self.indexdata.keys())[-1]

        try:
            entry_ID_numbers = [int(entryID.split(".")[-1]) for entryID in self.indexdata.keys()]

            #last_ID_num = int(last_entry_ID.split(".")[-1])
        except ValueError as ex:
            logger.error(f"Could not convert last part of an ID to an int")
            raise ManageError("internal-error", {})

        if len(entry_ID_numbers) > 0:
            last_ID_num = sorted(entry_ID_numbers)[-1]
        else:
            last_ID_num = 0

        if not IDprefix.endswith("."):
            IDprefix = IDprefix + "."

        return IDprefix + str(last_ID_num + 1)


    def setIndexEntry(self, ID:str, entry:MetadataIndexEntry):
        self.indexdata[ID] = entry

    def __enter__(self):

        self.storage.__enter__()

        self._load_index()

        return self

    def _write_index(self):
        
        # We never want to serialise Keys in any format other than without tag/properties
        current_key_serialise = Key.serialise_type
        Key.serialise_type = KeySerialiseType.NO_TAGS_PROPERTIES

        load_yaml.class_to_yaml_file(self, Path(self.storage.repodir).joinpath(self.index_filename))

        Key.serialise_type = current_key_serialise

    def persist(self):
        """ Writes any changes to the underlying storage layer """

        self._write_index()

    def __exit__(self, exc_type, exc_value, traceback):

        if exc_type is None:
            self._write_index()

        # This is going to commit any changes
        self.storage.__exit__(exc_type, exc_value, traceback)

    def _get_state(self):

        output = {}
        output["threatmodels"] = [value for value in self.indexdata.values()]

        return output

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node._get_state())

class ThreatModelMetaData:
    """
    Stores all ThreatModelVersionMetaData for all versions of a threat model

    Can retrieve entries, update new entries and persist all entries.  It also sets a version as the approved version.
    """

    def __init__(self, config:dict, storage:ThreatModelStorage, schemeID:str = None, location:str = None) -> None:
        yaml_register_class(ThreatModelMetaData)

        self.index = IndexMetaData(config, storage)

        self.metadata_filename = config.get("metadata-filename", "metadata.yaml")
        self.storage = storage
        self.approvedStatus = config.get("approved-status", "Approved")
        self.write_plain_model = config.get("write-plain-model", True)

        self.index._load_index()

        self.ID = self.index.getIDByLocation(location)
        if self.ID is None:
            logger.error(f"No ID was found in the metadata for a threat model with location '{location}'")
            raise ManageError("no-ID-in-metadata", {"location":location})

        self._set_empty()

        self._load_metadata() 

    def _set_empty(self):
        self.IDmetadata = {}
        self.IDmetadata["ID"] = self.ID
        self.IDmetadata["versions"] = {}

    def setModel(self, version:str, model:dict):
        self.model_version = version
        self.model = model

    def _load_metadata(self):

        # If the directory or file don't exist, that's fine as they'll be created when we write out changes.

        if (contents_dict := self.storage.load_yaml([], Path(self.ID).joinpath(self.metadata_filename))) is not None:

            if (tm_metadata := contents_dict.get("threatmodel-metadata", None)) is None:
                logger.error("Could not find key 'threatmodel-metadata'")
                return

            if not match.equals(self.ID, tm_metadata.get("ID", None)):
                logger.error("The 'ID' in the threatmodel metadata file did not match the ID of the threat model")
                return

            for version_key, version_value in tm_metadata.get("versions", {}).items():
                tmvm = ThreatModelVersionMetaData()
                tmvm.fromDict(version_key, version_value)
                self.IDmetadata["versions"][version_key] = tmvm

    def _write_metadata(self):
        
        # We never want to serialise Keys in any format other than without tag/properties
        current_key_serialise = Key.serialise_type
        Key.serialise_type = KeySerialiseType.NO_TAGS_PROPERTIES

        self.storage.write_yaml(Path(self.ID).joinpath(self.metadata_filename), self)

        Key.serialise_type = current_key_serialise
        

    def _write_model(self):
        """ Optionally, writes 2 versions of the model file, one that can be deserialised, and one that is human readable (plain) """
        Key.serialise_type = KeySerialiseType.TAGS_PROPERTIES
        #self.storage.write_yaml([Key], Path(self.ID).joinpath(self.model_version + ".yaml"), self.model)
        self.storage.write_yaml(Path(self.ID).joinpath(self.model_version + ".yaml"), self.model)

        if self.write_plain_model:
            Key.serialise_type = KeySerialiseType.NO_TAGS_PROPERTIES
            #self.storage.write_yaml([Key], Path(self.ID).joinpath(self.model_version + ".plain.yaml"), self.model)
            self.storage.write_yaml(Path(self.ID).joinpath(self.model_version + ".plain.yaml"), self.model)

    def load_model(self, version:str):

        Key.serialise_type = KeySerialiseType.TAGS_PROPERTIES
        self.model = self.storage.load_yaml([Key], Path(self.ID).joinpath(version + ".yaml"))
        if self.model is None:
            logger.error(f"Unable to load model for version '{version}'")
        else:
            self.model_version = version

        return self.model
            

    def setApprovedVersion(self, ID:str, current_version:str, currentModelIndexEntry:MetadataIndexEntry, approved_version:str, approvedModelIndexEntry:MetadataIndexEntry, approved_version_metadata:ThreatModelVersionMetaData):

        currentMetadataIndexEntry = self.index.getIndexEntry(ID)

        if currentMetadataIndexEntry is None:
            # The TM has either not been 'create'd, or the output of the 'create' action has not been merged to the default branch
            raise ManageError("cant-submit-not-in-metadata-index", {"ID":ID})

        # Several scenarios to consider
        # Common legitimate update scenario by author
        # model cur ver = 1.1 (not approved in model)
        # model app ver = 1.0 (approved in model)
        # index app ver = 1.0

        # Common legitimate update scenario by approver
        # model cur ver = 1.1 (approved in model)
        # model app ver = 1.1 (approved in model)
        # index app ver = 1.0

        # Common mistake update scenario by approver - forgot to update approved version in model
        # model cur ver = 1.1 (approved in model)
        # model app ver = 1.0 (approved in model)
        # index app ver = 1.0
    
        # Common mistake update scenario by author/approver - tried to update existing approved version
        # model cur ver = 1.1 (approved in model)
        # model app ver = 1.1 (approved in model)
        # index app ver = 1.1

        if currentModelIndexEntry.isApproved() and approvedModelIndexEntry.isApproved():
            if not match.equals(currentModelIndexEntry.approved_version, approvedModelIndexEntry.approved_version):
                # In the model, the current version is approved, and is different to the approved version from the model
                logger.error(f"Submitted Threat Model has different current and approved versions, but both are approved.  If the current version is approved it must match the approved version.")
                raise ManageError("current-approved-mismatched-version", {"ID":ID, "current_version":currentModelIndexEntry.approved_version, "approved_version":approvedModelIndexEntry.approved_version})
            elif match.equals(approvedModelIndexEntry.approved_version, currentMetadataIndexEntry.approved_version):
                # The model is correct, but we the index says the version is already approved
                logger.error(f"Submitted Threat Model approved version in the document is the same as the officially recorded approved version.  Submitting new versions requires a version increase.")
                raise ManageError("cant-update-approved-version", {"ID":ID, "approved_version":currentMetadataIndexEntry.approved_version})
        elif match.equals(current_version, approved_version) and not approvedModelIndexEntry.isApproved():
            # Often the only reason the TM is not considered approved is someone forgot to change the status
            if approvedModelIndexEntry.hasApproval() and not match.equals(self.approvedStatus, approvedModelIndexEntry.versionStatus()):
                logger.error(f"Submitted Threat Model has been approved, but status is not set to '{self.approvedStatus}'.")
                raise ManageError("status-not-approved", {"ID":ID, "approved_version":approved_version, "approved_status":self.approvedStatus})

            # The current version = approved version, but it's not approved.  Probably forgot to approve the version in the version history table
            logger.error(f"Submitted Threat Model approved version in the document is not marked as approved in the Version History table.")
            raise ManageError("approved-version-not-approved", {"ID":ID, "approved_version":approved_version})

        elif not match.equals(current_version, approved_version) and currentModelIndexEntry.isApproved():
            # The current version != approved version, but current version approved.  Probably forgot to increase the approved version the version in the version history table
            logger.error(f"Submitted Threat Model current version is approved, but differs from the approved version.")
            raise ManageError("approved-version-not-updated", {"ID":ID, "current_version":current_version, "approved_version":approved_version})


        self.index.setIndexEntry(ID, currentModelIndexEntry)

        self.IDmetadata["versions"][approved_version] = approved_version_metadata

    def persist(self):
        """ Writes any changes to the underlying storage layer """

        self.index._write_index()
        self._write_metadata()
        self._write_model()


    def getVersion(self, versionID:str) -> ThreatModelVersionMetaData:
        
        return self.IDmetadata.get(versionID, None)

    def _get_state(self):

        output = {}
        output["threatmodel-metadata"] = self.IDmetadata

        return output

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node._get_state())


import logging
import utils.logging

utils.logging.configureLogging()
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class Request:

    action:str
    scheme:str
    docloc:str
    doctemplate:str
    ID:str
    IDprefix:str
    lang:str
    format:str
    meta:str
    reports:str

    @classmethod
    def set(cls, request_parameters:dict) -> None:
        
        # Store the query string parameters
        cls.action = request_parameters.get("action", None)
        cls.scheme = request_parameters.get("scheme", None)
        cls.docloc = request_parameters.get("docloc", None)
        cls.doctemplate = request_parameters.get("doctemplate", None)
        cls.ID = request_parameters.get("ID", None)
        cls.IDprefix = request_parameters.get("IDprefix", None)
        cls.lang = request_parameters.get("lang", None)
        cls.format = request_parameters.get("format", "json")
        cls.meta = request_parameters.get("meta", "tags")
        cls.reports = request_parameters.get("reports", "none")

        logger.info(f"Threatware called with parameters = '{ request_parameters }'")

    @classmethod
    def get(cls) -> dict:
        return {
            "action": cls.action,
            "scheme": cls.scheme,
            "docloc": cls.docloc,
            "doctemplate": cls.doctemplate,
            "ID": cls.ID,
            "IDprefix": cls.IDprefix,
            "lang": cls.lang,
            "format": cls.format,
            "meta": cls.meta,
            "reports": cls.reports
        }
    
    @classmethod
    def isAPIFormat(cls) -> bool:
        return cls.format in ["json", "yaml"]


#!/usr/bin/env python3
"""
Translate class responsible for localisation
"""

import logging
from pathlib import Path
from utils import match
from utils.config import ConfigBase
from utils.load_yaml import yaml_file_to_str, yaml_str_to_dict, yaml_file_to_dict
from utils.request import Request

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

from jinja2 import Environment, FileSystemLoader, select_autoescape
env = Environment(
    loader = FileSystemLoader(searchpath="./"),
    autoescape=select_autoescape()
)

TRANSLATE_CONFIG_YAML = "translate.yaml"
TRANSLATE_CONFIG_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(TRANSLATE_CONFIG_YAML))

class Translate:

    translations:dict = {}
    global_context:dict = {}
    defLanguageCode:str = "default"
    defOutputFormatKey:str = "default"
    languageCode:str = ""
    _cache:dict = {}

    @classmethod
    def init(cls):
    #def init(cls, languageCode:str = "", global_context:dict = {}):

        #cls.global_context = global_context
        cls.global_context = {"request": Request.get()}
        languageCode = Request.lang

        yaml_config_dict = yaml_file_to_dict(ConfigBase.getConfigPath(TRANSLATE_CONFIG_YAML_PATH))

        if languageCode is None or languageCode == "":
            languageCode = cls.defLanguageCode

        if languageCode not in yaml_config_dict:
            logger.warning(f"Could not find language code '{languageCode}' in translation file '{TRANSLATE_CONFIG_YAML_PATH}'.  Using {cls.defLanguageCode}.")
            languageCode = cls.defLanguageCode
            if languageCode not in yaml_config_dict:
                logger.warning(f"Could not find language code '{languageCode}' in translation file '{TRANSLATE_CONFIG_YAML_PATH}'.")

        cls.languageCode = languageCode
        cls.translations = yaml_config_dict.get(cls.languageCode, {})

    @classmethod
    def localise(cls, texts:dict, texts_key:str = None, context:dict = {}, cache_key = None, ignore_format:bool = False):

        # localise is expensive to call a lot, so cache context free values.  This is fine as language does not change per execution
        if context is None or len(context) == 0:
            if cache_key is None:
                # TODO making texts hashable via str(texts) is one way, there might be faster other ways e.g. frozenset
                cache_key = str(texts)
            if (cached_value := cls._cache.get((cache_key, texts_key, ignore_format), None)) is not None:
                return cached_value

        if texts is None or len(texts) == 0:
            logger.error("Translate requires a dict of texts to localise from, but an empty dict was provided.")
            return ""

        # Get the language code (set in init)
        languageCode = cls.languageCode
        if languageCode not in texts:
            logger.warning(f"Could not find language code '{languageCode}' in text dict.  Using default.")
            if (languageCode := cls.defLanguageCode) not in texts:
                logger.warning(f"Could not find language code '{languageCode}' in text dict.")

        # Get the set of texts for that language
        textsLanguage = texts.get(languageCode, {texts_key:f"Could not find texts in language '{languageCode}'"})

        if not match.is_empty(texts_key):
            # textsLanguage is a dict
            # Lookup the texts_key in the set of texts for that language
            textsLanguageText = textsLanguage.get(texts_key, f"Could not find text for key '{texts_key}'")
        elif isinstance(textsLanguage, str):
            # textsLanguage is the actual string
            # Edge case where we want aren't looking for a key in a set of language texts, but rather 
            # have a set of language texts for a single key (which hasn't been passed in), so we want to 
            # just return the text associated with the language code
            textsLanguageText = textsLanguage
        else:
            # Since no texts_key has been passed in, and we can't translate an empty string, we'll just return an empty string
            # This can happen in error handling when an unexpected error occurs and we are trying to localise the error message
            logger.debug(f"No localisation text key passed in, so we can't choose a localised text from the supplied texts.  Returning an empty string.")
            return ""

        # We may have different versions of the text for different output formats e.g. json vs html
        if isinstance(textsLanguageText, dict):
            format = Request.format
            if ignore_format is True:
                format = "default"
            # Check if the requested output format text is available, otherwise use the default
            if format in textsLanguageText:
                textsLanguageText = textsLanguageText.get(format)
            elif cls.defOutputFormatKey in textsLanguageText:
                textsLanguageText = textsLanguageText.get(cls.defOutputFormatKey)
            else:
                logger.warning(f"The localised texts didn't have a format entry matching '{format}' or the default '{cls.defOutputFormatKey}'.")
                
        if isinstance(textsLanguageText, str):
            output = env.from_string(textsLanguageText).render(context | cls.translations | cls.global_context)
        elif isinstance(textsLanguageText, list):
            output = []
            for textsLanguageTextEntry in textsLanguageText:
                # Assume all entries are strings
                output.append(env.from_string(textsLanguageTextEntry).render(context | cls.translations | cls.global_context))
        else:
            logger.warning(f"Unsupported type '{type(textsLanguageText)}' passed to method.  Returning un-localised value.")
            output = textsLanguageText

        if context is None or len(context) == 0:
            cls._cache[(cache_key, texts_key, ignore_format)] = output

        return output

    @classmethod
    def localiseYamlFile(cls, filepath:Path) -> dict:

        yaml_str = yaml_file_to_str(filepath)

        localised_yaml_str = env.from_string(yaml_str).render(cls.translations | cls.global_context)

        return yaml_str_to_dict(localised_yaml_str)

    @classmethod
    def getTranslation(cls, key:str):
        return cls.translations.get("translate", {}).get(key, None)


#!/usr/bin/env python3
"""
Utility methods for transforming strings
"""

import logging

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def identity():
    return lambda val : val

def c14n(value:str, transform_fn = None) -> str:
    if transform_fn is not None:
        value = transform_fn(value)
    return value.casefold().strip()


def strip(start_char:str, end_char:str):
    """
    Strips everything between the start and end characters/str (inclusive)

    Returns a function.  Designed for use with methods taking transform_fn parameters that end up calling c14n e.g. util.match
    """
    def _strip_context(value:str,start_char:str, end_char:str):

        if(start_char_index := value.find(start_char)) != -1:
            if(end_char_index := value.find(end_char, start_char_index)) != -1:
                return value[:start_char_index] + value[end_char_index + 1:]

        return value

    return lambda val : _strip_context(val, start_char,  end_char)

